import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from collections import defaultdict
import copy as copy
import warnings
import time

from azapy.MkT.MkTcalendar import NYSEgen

_WW_TOL_ = 1.e-10
_MU_ABS_TOL_ = 1.e-6
_RISK_REL_TOL_ = 0.999999
_RR_ABS_TOL_ = 1.e-4


class _RiskAnalyzer:
    """
    Base class for all XXXAnalyzer classes. \n
    All derive classes need to implement: \n
        _risk_calc(self, prate, alpha) \n
        _risk_min(self, d=1) \n
        _sharpe_max(self) \n
        _sharpe_inv_min(self) \n
        _rr_max(self) \n
        _risk_averse(self) \n
        _risk_diversification(self, d=1) \n
        _risk_inv_diversification(self, d=1) \n
        _rr_max_diversification(self) \n
        _set_method(self, method) \n

    Methods:
        * getWeights
        * getRisk
        * getPositions
        * getRiskComp
        * getDiversification
        * viewForntiers
        * set_rrate
        * set_method
        * set_mktdata
        * set_rtype
        * set_random_seed
        
    Public members:
        
    """

    def __init__(self, mktdata=None, colname='adjusted', freq='Q', 
                 hlength=3.25, calendar=None, rtype='Sharpe', name=None):
        """
        Constructor

        Parameters
        ----------
        `mktdata` : `pandas.DataFrame`, optional;
            Historic daily market data for portfolio components in the format
            returned by `azapy.mktData` function. The default is `None`.
        `colname` : `str`, optional;
            Name of the price column from mktdata used in the weights
            calibration. The default is `'adjusted'`.
        `freq` : `str`, optional;
            Rate of return horizon in number of business day. It could be
            `'Q'` for quarter or `'M'` for month. The default is `'Q'`.
        `hlength` : `float`, optional;
            History length in number of years used for calibration. A
            fractional number will be rounded to an integer number of months.
            The default is `3.25` years.
        `calendar` : `numpy.busdaycalendar`, optional;
            Business days calendar. If is it `None` then the calendar will be 
            set to NYSE business calendar. The default is `None`.
        `rtype` : `str`, optional;
            Optimization type. Possible values: \n
                `'Risk'` : optimal risk portfolio for targeted expected rate of 
                return.\n
                `'Sharpe'` : optimal Sharpe portfolio - maximization solution.\n
                `'Sharpe2'` : optimal Sharpe portfolio - minimization solution.\n
                `'MinRisk'` : minimum risk portfolio.\n
                `'RiskAverse'` : optimal risk portfolio for a fixed 
                risk-aversion factor.\n
                `'InvNrisk'` : optimal risk portfolio with the same risk value 
                as a benchmark portfolio (e.g. same as equal weighted 
                portfolio).\n
                `'Diverse'` : optimal diversified portfolio for targeted
                expected rate of return (max of inverse 1-Diverse).\n
                `'Diverse2'` : optimal diversified portfolio for targeted
                expected rate of return (min of 1-Diverse).\n
                `'MaxDiverse'` : maximum diversified portfolio.\n
                `'InvNdiverse'` : optimal diversified portfolio with the same
                diversification factor as a benchmark portfolio 
                (e.g. same as equal weighted portfolio).\n
                `'InvNdrr'` : optimal diversified portfolio with the same 
                expected rate of return as a benchmark portfolio
                (e.g. same as equal weighted portfolio).\n
            The default is `'Sharpe'`.
        `name` : `str`, optional;
            Object name. Deafult value is `None`.

        Returns
        -------
        The object.
        """
        self.ww = None
        self.status = None

        self.last_data = None
        self.rrate = None
        self.nn = None
        self.mm = None
        self.muk = None
        self.mu = None
        
        self.alpha = [1.]
        self.coef = [1.]
        self.Lambda = None

        self.risk = None
        self.primary_risk_comp = None
        self.secondary_risk_comp = None
        self.sharpe = None
        self.RR = None
        self.diverse = None
        self.time_level1 = None
        self.time_level2 = None
        self.time_level3 = None

        self.rtype = None
        self.rng = None
        self.method = None
        
        self.methods = None
        self.rtypes = None
        self.name = None
        
        self.risk_comp = None
        self._flag_risk_comp_calc = False
        self._status_risk_comp_calc = None
        
        self.set_rtype(rtype)
        
        self.name = name

        if mktdata is not None:
            self.set_mktdata(mktdata, colname, freq, hlength)
        self.set_random_seed()
        
        self.verbose = False
        
        
    def _reset_output(self):
        self.status = None
        self.ww = None
        self.risk = None
        self.primary_risk_comp = None
        self.secondary_risk_comp = None
        self.sharpe = None
        self.RR = None
        self.diverse = None
        self.time_level1 = None
        self.time_level2 = None
        self.time_level3 = None

        
    def getWeights(self, rtype=None, mu=None, d=1, mu0=0., aversion=None,
                   ww0=None, rrate=None, verbose=False):  
        """
        Computes the optimal portfolio weights.

        Parameters
        ----------
        `rtype` : `str`, optional;
            Optimization type. If is not `None` it will overwrite the value
            set by the constructor. The default is `None`.
            Other possible values: \n
                `'Risk'` : optimal risk portfolio for targeted expected rate of 
                return.\n
                `'Sharpe'` : optimal Sharpe portfolio - maximization solution.\n
                `'Sharpe2'` : optimal Sharpe portfolio - minimization solution.\n
                `'MinRisk'` : minimum risk portfolio.\n
                `'RiskAverse'` : optimal risk portfolio for a fixed 
                risk-aversion factor.\n
                `'InvNrisk'` : optimal risk portfolio with the same risk value 
                as a benchmark portfolio (e.g. same as equal weighted 
                portfolio).\n
                `'Diverse'` : optimal diversified portfolio for targeted
                expected rate of return (max of inverse 1-Diverse).\n
                `'Diverse2'` : optimal diversified portfolio for targeted
                expected rate of return (min of 1-Diverse).\n
                `'MaxDiverse'` : maximum diversified portfolio.\n
                `'InvNdiverse'` : optimal diversified portfolio with the same
                diversification factor as a benchmark portfolio 
                (e.g. same as equal weighted portfolio).\n
                `'InvNdrr'` : optimal diversified portfolio with the same 
                expected rate of return as a benchmark portfolio
                (e.g. same as equal weighted portfolio).\n
        `mu` : `float`, optional;
            Targeted portfolio expected rate of return. 
            Relevant only if `rtype` is equalt to `'Risk'` or `'Divers'`.
            The default is `None`.
        `d` : `int`, optional;
            Frontier type. Active only if `rtype='Risk'`. A value of `1` will
            trigger the evaluation of optimal portfolio along the efficient
            frontier. Otherwise, it will find the portfolio with the lowest
            rate of return along the inefficient portfolio frontier.
            The default is `1`.
        `mu0` : `float`, optional;
            Risk-free rate accessible to the investor.
            Relevant only if `rype='Sharpe'` or `rtype='Sharpe2'`.
            The default is `0`.
        `aversion` : `float`, optional;
            The value of the risk-aversion coefficient.
            Must be positive. Relevant only if `rtype='RiskAverse'`.
            The default is `None`.
        `ww0` : `list`, `numpy.array` or `pandas.Series`, optional;
            Targeted portfolio weights. 
            Relevant only if `rype='InvNrisk'`.
            Its length must be equal to the number of symbols in `rrate` 
            (mktdata). All weights must be >= 0 with sum > 0.
            If it is a `list` or a `numpy.array` then the weights are assumed 
            to be in order of `rrate.columns`. If it is a `pandas.Series` then 
            the index should be compatible with the `rrate.columns` or mktdata 
            symbols (same symbols, not necessary in the same order).
            If it is `None` then it will be set to equal weights.
            The default is `None`.
        `rrate` : `pandas.DataFrame`, optional;
            The portfolio components historical rates of returns.
            If it is not `None`, it will overwrite the `rrate` computed in the
            constructor from mktdata. The default is `None`. 
        `verbose` : Boolean, optiona;
            If it set to `True` then it will print messages when the optimal
            portfolio degenerates to a single asset portfolio as a limited 
            case. 
            The default is `False`.

       Returns
        -------
        `pandas.Series` with portfolio weights.
        """
        toc = time.perf_counter()
        self._reset_output()
        self.verbose = verbose
 
        d = 1 if d == 1 else -1

        if rrate is not None:
            self.set_rrate(rrate)
            
        if not rtype is None:
            self.set_rtype(rtype)
            
        # calc - grand dispatch 
        if self.rtype == "Risk":
            self._calc_Risk(mu, d)
        elif self.rtype == 'Sharpe':           
            self._calc_Sharpe(mu0)
        elif self.rtype == 'Sharpe2':          
            self._calc_Sharpe2(mu0)
        elif self.rtype == "MinRisk":
            self._calc_MinRisk()
        elif self.rtype == "InvNrisk":    
            self._calc_InvNrisk(ww0)
        elif self.rtype == "RiskAverse":   
            self._calc_RiskAverse(aversion) 
        elif self.rtype == "Diverse2":
            self._calc_Diverse2(mu, d)
        elif self.rtype == "Diverse":
            self._calc_Diverse(mu, d)
        elif self.rtype == "MaxDiverse":  
            self._calc_MaxDiverse()
        elif self.rtype == "InvNdiverse":
            self._calc_InvNdiverse(ww0)
        elif self.rtype == "InvNdrr":
            self._calc_InvNdrr(ww0)
        else:
            raise ValueError("you should not be here - ever")

        if self.status != 0:
            warnings.warn(f"Warning: {self.rtype} on {self.rrate.index[-1]} "
                          f"status {self.status}")
        else:
            self.ww = pd.Series(self.ww, index=self.rrate.columns)
            self._norm_weights()
            
        self.time_level1 = time.perf_counter() - toc
        
        return self.ww
    
    
    def _calc_Risk(self, mu, d):
        mu_max = self.muk.max()
        mu_min = self.muk.min()
        if mu is None:
            raise ValueError("for rtype='Risk' mu must have a value")
        elif d == 1 and mu - mu_max > -np.abs(mu_max) * _MU_ABS_TOL_:
            self._single_asset_port('max')
            if self.verbose:
                print(f"rtype 'Risk' for mu {mu} for {self.name} on "
                      f"{self.rrate.index[-1]}, defaulted "
                      f"to single portfolio [{self.muk.idxmax()}]")
        elif d == -1 and mu - mu_min < np.abs(mu_min) * _MU_ABS_TOL_:
            self._single_asset_port('min')
            if self.verbose:
                print(f"rtype 'Risk' for mu {mu} for {self.name} on "
                      f"{self.rrate.index[-1]}, defaulted "
                      f"to single portfolio [{self.muk.idxmin()}]")
        else:
            self.mu = mu
            self._risk_min(d=d)
            
            
    def _calc_Sharpe(self, mu0):
        mu_max = self.muk.max()
        if mu0 is None:
            raise ValueError("for rtype='Sharpe' mu0 must have a value")
        if mu0 - mu_max >= -np.abs(mu_max) * _MU_ABS_TOL_:
            self._single_asset_port('max')
            if self.status == 0:
                self.sharpe = (self.RR - mu0) / self.risk
                if self.verbose:
                    print(f"rtype 'Sharpe' for mu0 {mu0} for {self.name} on "
                          f"{self.rrate.index[-1]}, defaulted "
                          f"to single portfolio [{self.muk.idxmax()}]")
        else:
            self.mu = mu0
            self._sharpe_max()
            
            
    def _calc_Sharpe2(self, mu0):
        mu_max = self.muk.max()
        if mu0 is None:
            raise ValueError("for rtype='Sharpe' mu0 must have a value")
        if mu0 - mu_max >= -np.abs(mu_max) * _MU_ABS_TOL_:
            self._single_asset_port('max')
            if self.status == 0:
                self.sharpe = (self.RR - mu0) / self.risk
                if self.verbose:
                    print(f"rtype 'Sharpe2' for mu0 {mu0} for {self.name} on "
                          f"{self.rrate.index[-1]}, defaulted "
                          f"to single portfolio [{self.muk.idxmax()}]")
        else:
            self.mu = mu0
            self._sharpe_inv_min() 
            
            
    def _calc_MinRisk(self):
        self._calc_Risk(self.muk.min(), d=1)
        
        
    def _calc_InvNrisk(self, ww0):
        if ww0 is None:
            ww = np.full(len(self.rrate.columns), 
                         1/len(self.rrate.columns))
        else:
            if isinstance(ww0, pd.core.series.Series):
                ww = np.array(ww0[self.rrate.columns])
            else:
                ww = np.array(ww0)
            if any(ww < 0.):
                raise ValueError("All ww0 must be non-negative")
            sww = ww.sum()
            if sww <= 0:
                raise ValueError("At least one ww0 must be non zero")
            ww = ww / sww
            
        self.getRiskComp()
        if self.status == 0:
            self.getRisk(ww)
            if self.status == 0:
                symb_max = self.muk.idxmax()
                if self.risk < self.risk_comp[symb_max] * _RISK_REL_TOL_:
                    self._rr_max()  
                else:
                    w = pd.Series(0., index=self.rrate.columns)
                    w[symb_max] = 1.
                    self.getRisk(w)
                    if self.verbose:
                        print(f"InvNrisk for {self.name} on "
                              f"{self.rrate.index[-1]}, defaulted "
                              f"to single portfolio [{symb_max}]")
                    
                    
    def _calc_RiskAverse(self, aversion):
        if aversion is None:
            raise ValueError("for rtype='RiskAverse'"
                             " aversion must have a positive value")
        elif aversion < 0:
            raise ValueError("aversion must be positive (>0)")
            
        self.Lambda = aversion
        self._risk_averse()
        
        
    def _calc_Diverse(self, mu, d):
        mu_max = self.muk.max()
        mu_min = self.muk.min()
        if mu is None:
            raise ValueError("for rtype='Diverse' mu must have a value")
        elif d == 1 and mu - mu_max > -np.abs(mu_max) * _MU_ABS_TOL_:
            self._single_asset_port('max')
            if self.verbose:
                print(f"rtype 'Diverse' for mu {mu} for {self.name} on "
                      f"{self.rrate.index[-1]}, defaulted "
                      f"to single portfolio [{self.muk.idxmax()}]")
        elif d == -1 and mu - mu_min < np.abs(mu_min) * _MU_ABS_TOL_:
            self._single_asset_port('min')
            if self.verbose:
                print(f"rtype 'Diverse' for mu {mu} for {self.name} on "
                      f"{self.rrate.index[-1]}, defaulted "
                      f"to single portfolio [{self.muk.idxmin()}]")
        else:
            self.mu = mu
            self.getRiskComp()
            if self.status == 0:
                self._risk_inv_diversification(d=d)
                
                
    def _calc_Diverse2(self, mu, d):
        mu_max = self.muk.max()
        mu_min = self.muk.min()
        if mu is None:
            raise ValueError("for rtype='Diverse2' mu must have a value")
        elif d == 1 and mu - mu_max > -np.abs(mu_max) * _MU_ABS_TOL_:
            self._single_asset_port('max')
            if self.verbose:
                print(f"rtype 'Diverse2' for mu {mu} for {self.name} on "
                      f"{self.rrate.index[-1]}, defaulted "
                      f"to single portfolio [{self.muk.idxmax()}]")
        elif d == -1 and mu - mu_min < np.abs(mu_min) * _MU_ABS_TOL_:
            self._single_asset_port('min')
            if self.verbose:
                print(f"rtype 'Diverse2' for mu {mu} for {self.name} on "
                      f"{self.rrate.index[-1]}, defaulted "
                      f"to single portfolio [{self.muk.idxmin()}]")
        else:
            self.mu = mu
            self.getRiskComp()
            if self.status == 0:
                self._risk_diversification(d=d)
                
                
    def _calc_MaxDiverse(self):
        self._calc_Diverse(self.muk.min(), d=1)
        
        
    def _calc_InvNdiverse(self, ww0):
        if ww0 is None:
            ww = np.full(len(self.rrate.columns), 
                         1/len(self.rrate.columns))
        else:
            if isinstance(ww0, pd.core.series.Series):
                ww = np.array(ww0[self.rrate.columns])
            else:
                ww = np.array(ww0)
            if any(ww < 0.):
                raise ValueError("All ww0 must be non-negative")
            sww = ww.sum()
            if sww <= 0:
                raise ValueError("At least one ww0 must be non zero")
            ww = ww / sww
            
        self.getDiversification(ww)
        if self.status == 0:
            self._rr_max_diversification()  
            
            
    def _calc_InvNdrr(self, ww0):
        if ww0 is None:
            ww = np.full(len(self.rrate.columns), 
                         1/len(self.rrate.columns))
        else:
            if isinstance(ww0, pd.core.series.Series):
                ww = np.array(ww0[self.rrate.columns])
            else:
                ww = np.array(ww0)
            if any(ww < 0.):
                raise ValueError("All ww0 must be non-negative")
            sww = ww.sum()
            if sww <= 0:
                raise ValueError("At least one ww0 must be non zero")
            ww = ww / sww
            
        mu = np.dot(ww, self.muk)
        self._calc_MaxDiverse()
        if self.status == 0:
            if mu > self.RR:
                self._calc_Diverse(mu, d=1)
            elif mu < self.RR:
                self._calc_Diverse(mu, d=-1)
        

    def getRisk(self, ww, rrate=None):
        """
        Returns the value of the dispersion (risk) measure for a give 
        portfolio.

        Parameters
        ----------
        `ww` : `list`, `numpy.array` or `pandas.Series`;
            Portfolio weights. Its length must be equal to the number of
            symbols in `rrate` (mktdata). All weights must be >=0 with 
            sum > 0.
            If it is a `list` or a `numpy.array` then the weights are assumed 
            to be in order of `rrate.columns`. If it is a `pandas.Series` than 
            the index should be compatible with the `rrate.columns` or mktdata 
            symbols (not necessary in the same order).
        `rrate` : `pandas.DataFrame`, optional;
            Contains the portfolio components historical
            rates of returns. If it is not `None`, it will overwrite the
            `rrate` computed in the constructor from `mktdata`.
            The default is `None`.

        Returns
        -------
        `float`
            The dispersion (risk) measure value.
        """
        self._reset_output()
        toc = time.perf_counter()
        
        if rrate is not None:
            self.set_rrate(rrate)

        if isinstance(ww, pd.core.series.Series):
            ww = ww[self.rrate.columns]
        w = np.array(ww)
        if any(w < 0.):
            raise ValueError("All ww must be non-negative")
        sw = w.sum()
        if sw <= 0:
            raise ValueError("At least one ww must be positive")
        w = w / sw

        prate = np.dot(self.rrate, w)

        self.primary_risk_comp = []
        self.secondary_risk_comp = []
        for alpha in self.alpha:
            self.status, secondary, primary = self._risk_calc(prate, alpha)
            if self.status != 0:
                warnings.warn(f"status: {self.status}, \
                              wrong risk calc for alpha {alpha}")
                return np.nan

            self.primary_risk_comp.append(primary)
            self.secondary_risk_comp.append(secondary)
        self.risk = np.dot(self.primary_risk_comp, self.coef)
        self.primary_risk_comp = np.array(self.primary_risk_comp)
        self.secondary_risk_comp = np.array(self.secondary_risk_comp)
        self.RR = np.dot(w, self.muk)
        self.ww = w
        self.time_level1 = time.perf_counter() - toc
        
        return self.risk

        
    def getPositions(self, nshares=None, cash=0, ww=None, rtype=None, 
                     mu=None, mu0=0., aversion=None, ww0=None, ):
        """
        Computes the rebalanced number of shares.

        Parameters
        ----------
        `nshares` : `panda.Series`, optional;
            Initial number of shares per portfolio component.
            A missing component
            entry will be considered 0. A `None` value assumes that all
            components entries are 0. The name of the components must be
            present in the mrkdata. The default is `None`.
        `cash` : `float`, optional;
            Additional cash to be added to the capital. A
            negative entry assumes a reduction in the total capital
            available for rebalance. The total capital cannot be < 0.
            The default is 0. 
        `ww` : `panda.Series`, optional;
            External overwrite portfolio weights. 
            If it not set to `None` these
            weights will overwrite the calibration results.
            The default is `None`. 
        `rtype` : `str`, optional;
            Optimization type. If is not `None` it will overwrite the value
            set by the constructor. The default is `None`.
        `mu` : `float`, optional
            Targeted portfolio expected rate of return. 
            Relevant only if `rtype='Risk'` and `rtype='Divers'`.
            The default is `None`.
        `mu0` : `float`, optional;
            Risk-free rate accessible to the investor.
            Relevant only if `rype='Sharpe'` or `rtype='Sharpe2'`.
            The default is `0`.
        `aversion` : `float`, optional;
            The value of the risk-aversion coefficient.
            Must be positive. Relevant only if `rtype='RiskAvers'`.
            The default is `None`.
        `ww0` : list (also `numpy.array` or `pandas.Series`), optional;
            Targeted portfolio weights 
            Relevant only if `rype='InvNrisk'`.
            Its length must be equal to the number of
            symbols in `rrate` (mktdata). 
            All weights must be >= 0 with sum > 0.
            If it is a `list` or a `numpy.array` then the weights are assumed to
            be in order of `rrate.columns`. If it is a `pandas.Series` then the 
            index should be compatible with the `rrate.columns` or mktdata 
            symbols (same symbols, not necessary in the same order).
            If it is `None` then it will be set to equal weights.
            The default is `None`.

        Returns
        -------
        `pandas.DataFrame`: the rolling information.

        Columns:

            - `"old_nsh"` :
                initial number of shares per portfolio component and
                the additional cash. These are input values.
            - `"new_nsh"` :
                the new number of shares per component plus the residual
                cash (due to the rounding to an integer number of shares).
                A negative entry means that the investor needs to add more
                cash to cover for the roundup shortfall.
                It has a small value.
            - `"diff_nsh"` :
                the number of shares that needs to be both/sold in order
                to rebalance the portfolio positions.
            - `"weights"` :
                portfolio weights used for rebalancing. The cash entry is
                the new portfolio value (invested capital).
            - `"prices"` :
                the share prices used for rebalances evaluations.

        Note: Since the prices are closing prices, the rebalance can be
        computed after the market close and before the 
        trading execution (next day). 
        Additional cash slippage may occur due
        to share price differential between the previous day closing and
        execution time.
        """
       
        if rtype is not None:
            self.set_rtype(rtype)

        ns = pd.Series(0, index=self.rrate.columns)
        if nshares is not None:
            ns = ns.add(nshares, fill_value=0)

        if len(self.rrate.columns) != len(ns):
            raise ValueError("nshares must be a subset " + \
                             f"of {self.rrate.columns}")

        pp = self.last_data[self.rrate.columns.to_list()]

        cap = ns.dot(pp) + cash
        if ww is None:
            ww = self.getWeights(rtype=rtype, mu=mu, mu0=mu0, ww0=ww0, 
                                 aversion=aversion)
        else:
            wwp = pd.Series(0, index=self.rrate.columns)
            ww = wwp.add(ww, fill_value=0)
            if len(self.rrate.columns) == len(ns):
                raise ValueError(f"ww must be among {self.rrate.columns}")

        newns = (ww / pp * cap).round(0)
        newcash = cap - newns.dot(pp)

        ns['cash'] = cash
        newns['cash'] = newcash
        ww['cash'] = cap - newcash
        pp['cash'] = 1.

        res = pd.DataFrame({'old_nsh': ns,
                            'new_nsh': newns,
                            'diff_nsh': newns - ns,
                            'weights': ww.round(6),
                            'prices': pp})

        return res


    def set_rrate(self, rrate):
        """
        Sets portfolio components historical rates of returns.
        It will overwrite the value computed by the constructor from mktdata.

        Parameters
        ----------
        `rrate` : `pandas.DataFrame`;
            Portfolio components historical rates of returns. The
            columns are: "date", "symbol1", "symbol2", etc.
            
        Returns
        -------
        None
        """
        self._flag_risk_comp_calc = False
        self.nn, self.mm = rrate.shape
        self.muk = rrate.mean()
        self.rrate = rrate - self.muk


    def set_mktdata(self, mktdata, colname='adjusted',
                    freq='Q', hlength=3.25, calendar=None):
        """
        Sets historical market data. It will overwrite the choice made in the
        constructor.

        Parameters
        ----------
        `mktdata` : `pandas.DataFrame`;
            Historic daily market data for portfolio components in the format
            returned by `azapy.mktData` function.
        `colname` : `str`, optional;
            Name of the price column from mktdata used in the weights
            calibration. The default is 'adjusted'.
        `freq` : `str`, optional;
            Rate of returns horizon in number of business day. It could be
            'Q' for quarter or 'M' for month. The default is 'Q'.
        `hlength` : `float`, optional;
            History length in number of years used for calibration. A
            fractional number will be rounded to an integer number of months.
            The default is `3.25`.
        `calendar` : `numpy.busdaycalendar`, optional;
            Business days calendar. If is it `None` then the calendar will be 
            set to NYSE business calendar.
            The default is `None`.

        Returns
        -------
        `None`
        """
        if mktdata is None:
            return

        self.mktdata = mktdata

        periods = 63 if freq == 'Q' else 21

        if calendar is None:
            calendar = NYSEgen()

        edate = mktdata.index[-1]
        sdate = edate - pd.offsets.DateOffset(months=round(hlength * 12, 0))
        sdate = np.busday_offset(sdate.date(),
                                 0, roll='backward',  busdaycal=calendar)
        rrate = mktdata.pivot(columns='symbol', values=colname)[sdate:]
        self.last_data = rrate.iloc[-1]
        rrate = rrate.pct_change(periods=periods).dropna()

        self.set_rrate(rrate)


    def set_rtype(self, rtype):
        """
        Sets the optimization type. It will overwrite the value set in the
        constructor.

        Parameters
        ----------
        `rtype` : `str`;
            Optimization type.
            
        Returns
        -------
        `None`
        """
        self.rtypes = ['Risk', 'MinRisk', 'Sharpe', 'Sharpe2', 'RiskAverse',
                       'InvNrisk', 'Diverse', 'Diverse2', 'MaxDiverse', 
                       'InvNdiverse', 'InvNdrr']

        if not rtype in self.rtypes:
            raise ValueError(f"rtype (value {rtype}) must be "
                             f"one of {self.rtypes}")
        self.rtype = rtype


    def viewFrontiers(self, minrisk=True, efficient=20, inefficient=20, 
                      sharpe=True, musharpe=0., maxdiverse=False, 
                      diverse_efficient=0, diverse_inefficient=0, invN=True, 
                      invNrisk=True, invNdiverse=False, invNdrr=False,
                      component=True, randomport=20, addport=None, 
                      fig_type='RR_risk', **opt):
        """
        Computes the elements of the portfolio frontiers.

        Parameters
        ----------
        `minrisk` : Boolean, optional;
            If it is `True` then the minimum risk portfolio will be visible. 
            The default is `True`.
        `efficient` : `int`, optional;
            Number of points along the efficient risk frontier 
            (equally spaced along the rate of return axis). 
            The default is `20`.
        `inefficient` : `int`, optional;
            Number of points along the inefficient risk frontier 
            (equally spaced along the rate of returns axis). 
            The default is `20`.
        `sharpe` : Boolean, optional;
            If it is `True` then the maximum Sharpe portfolio will be visible. 
            The default is `True`.
        `musharpe` : `float`, optional;
            Risk-free rate value used in the evaluation of
            generalized Sharpe ratio. The default is `0`.
        `maxdiverse`: Boolean, optional;
            If it is `True` then the maximum diversified portfolio will be 
            visible. The default is `True`.
        `diverse_efficient`: `int`, optional;
            Number of points along the efficient diversification frontier
            (equally spaced along the rate of return axis).
            The default is `20`.
        `diverse_inefficient`: `int`, optional;
            Number of points along the inefficient diversification frontier
            (equally spaced along the rate of return axis).
            The default is `20`.
        `invN` : Boolean, optional;
            If it is `True`, then the equal weighted portfolio and the optimal 
            portfolio with the same risk value are added to
            the plot. The default is `True`.
        `invNrisk` : Boolena, optional;
            If it is `True`, then the efficient risk portfolio with same risk 
            as equal weighth portfolio is added to the plot.
            The defualt is `False`.
        `invNdiverse` : Boolean, optional:
            If it is `True`, then the efficient diversified portfolio with the 
            same diversification factor as the equal weighted portfolio is
            added to the plot. The default is `False`.
        `invNdrr` : Boolena, optional;
            If it is `True`, then the efficient diversified portfolio with the 
            same expected rate of return as the equal weighted portfolio is
            added to the plot. The default value is `False`.
        `component` : Boolean, optional;
            If it is `True`, then the single component portfolios are added to 
            the plot. The default is `True`.
        `randomport` : `int`, optional;
            Number of portfolios with random weights (inefficient) to be
            added to the plot (for reference). The default is `20`.
        `addport` : `dict` or `pandas.DataFrame`, optional;
            The weights of additional portfolio to be added to the plot.
            If it is a `dict` then the keys are the labels, and the values are 
            list of weights in the order of `rrate.columns`. If it is a 
            `pandas.DataFrame` the index are the labels, and each row is a set 
            of weights. The columns names should match the symbols names. 
            The default is `None`.
        `fig_type` : `str`, optional;
            Graphical representation format. \n
                * `'RR_risk'` : expected rate of return vs risk
                * `'Sharpe_RR'` : sharpe vs expected rate of return
                * `'Diverse_RR'` : diversification vs expected rate of return
                
            The default is `'RR_risk'`.
        `opt` : optional;
            Additonal parameters:\n
                * `'title'` : `str`; 
                    The default is 'Portfolio frontiers'
                * `'xlabel'` : `str`; 
                    The default is \n
                    - `'risk'` if `fig_type='RR_risk'`,
                    - `'rate of returns'` otherwise.
                * `'ylabel'` : `str`; 
                    The default is \n
                        - `'rate of returns'` if `fig_type='RR_risk'`
                        - `'sharpe'` if `fig_type='Sharpe_RR'`
                        - `'diversification'` if `fig_type=diverse_RR`   
                * `'tangent'` : Boolean; 
                    If set to `True`, then the tangent
                    (to max sharpe point) is added. It has effect only  if
                    `fig_type='RR_risk'`. The default is `True`.
                * `saveto` : `str`; 
                    File name to save the figure. The extension dictates the 
                    format: png, pdf, svg, etc. For more details see the 
                    `mathplotlib` documentation for `savefig`. 
                    The default is `None`.
                * `data` : `defaultdict`; 
                    Numerical data to construct the plot. If it is not `None`,
                    then it will take precedence and no other numerical 
                    evaluations will be performed. 
                    It is meant to produce different plot representations
                    without reevaluations. The default is `None`.
                * `invN_label` : `str`;
                    The label for equal weighted portfolio. 
                    The default is `'1/N'`.
                * `invNrisk_label` : `str`;
                    The lable for efficient portfolio with same risk as 
                    equal weighted portfolio. The default is `invNrisk`.
                * `invNdiverse_label` : `str`;
                    The label for diverse-efficient portfolio with the same
                    diversificeation factor as equal weighted porfolio.
                    The default is `'invNdiv'`.
                * `invNdrr_label` : `str`;
                    The label of diverse-efficient portfolio with the same
                    expected rate of return as equal weighted portolio.
                    The defualt is `'invNdrr'`.
                * `maxdiverse_label` : `str`;
                    The label of maximum diverified portfolio. 
                    The default is `'MaxD'`.
                * `minrisk_label` : `'str'`;
                    The label of minimum risk portfolio.
                    The default is `'MinR'`.
                * `sharpe_label` : `str`;
                    The label of maximum Sharpe portfolio. 
                    The default is `'Sharpe'`.

        Returns
        -------
        `dict`
            Numerical data used to make the plots. It can be passed back to
            reconstruct the plots without reevaluations.
        """
        if fig_type == 'RR_risk':
            xlabel = 'risk'
            ylabel = 'rate of return'
            plot_ff = self._plot_f1
        elif fig_type == 'Sharpe_RR':
            xlabel = 'rate of return'
            ylabel = 'sharpe'
            plot_ff = self._plot_f2
        elif fig_type == 'Diverse_RR':
            xlabel = 'rate of return'
            ylabel = 'diversification'
            plot_ff = self._plot_f3
        else:
            fty = ['RR_risk', 'Sharpe_RR', 'Diverse_RR']
            raise ValueError(f"fig_type={fig_type} ; must be one of {fty}")
                
        options_default = {'invN_label': '1/N',
                           'invNrisk_label': 'InvNrisk',
                           'invNdiverse_label' : 'InvNdiv',
                           'invNdrr_label' : 'InvNdrr',
                           'maxdiverse_label' : 'MaxD',
                           'minrisk_label' : 'MinR',
                           'sharpe_label' : 'Sharpe',
                           'comp_label': True,
                           'addport_label': True,
                           'tangent' : True,
                           'title' : 'Portfolio frontiers',
                           'xlabel' : xlabel,
                           'ylabel' : ylabel,
                           }
        options = defaultdict(lambda: None, options_default)
        for k, v in opt.items():
            if isinstance(v, dict) & (k != 'data'):
                options.update(v)
            else:
                options[k] = v

        resout = options['data']
        del options['data']
 
        # calc
        if resout is None:
            res = defaultdict(lambda: defaultdict(lambda: None))
            
            minrisk_calc = False
            if minrisk or efficient != 0 or inefficient != 0:
                minrisk_calc = True
            
            maxdiverse_calc = False
            self.getRiskComp()
            if maxdiverse or diverse_efficient != 0 or diverse_inefficient != 0: 
                maxdiverse_calc = True
  
            # min risk
            res['risk_min']['risk_min'] = minrisk
            if minrisk_calc:
                ww_min = self.getWeights(rtype='MinRisk')
                if self.status != 0:
                    res['risk_min']['risk_min'] = False
                    warnings.warn("Warning: MinRisk failed")
                else:
                    risk_min = self.risk
                    rr_min = self.RR
                    res['risk_min']['risk'] = risk_min
                    res['risk_min']['rr'] = rr_min
                    res['risk_min']['ww'] = ww_min
                    res['risk_min']['sharpe'] = \
                        (np.dot(ww_min, self.muk) - musharpe) / risk_min \
                            if risk_min != 0 else 0
                    res['risk_min']['diverse'] = self.getDiversification(ww_min)
                    res['risk_min']['label'] = options['minrisk_label']
    
            # efficient frontier
            res['efficient']['efficient'] = 0
            if efficient > 0:
                if rr_min is None:
                    warnings.warn("Warning: MinRisk -> efficient failed")
                    res['efficient']['efficient'] = 0
                else:
                    mu_max = self.muk.max()
                    rr_upper = mu_max - np.abs(mu_max) * _RR_ABS_TOL_
                    rr_eff = np.linspace(rr_min, rr_upper, efficient)
                    risk_eff = []
                    ww_eff = []
                    RR_eff = []
                    sharpe_eff = []
                    diverse_eff = []
                    for rr in rr_eff:
                        _ww = self.getWeights(rtype='Risk', mu=rr)
                        if self.status != 0:
                            warnings.warn(f"Warning: efficient for rr {rr} "
                                          "failed")
                            continue
                        ww_eff.append(_ww)
                        risk_eff.append(self.risk)
                        RR_eff.append(self.RR)
                        sharpe_eff.append(
                            (np.dot(self.ww, self.muk) - musharpe) / self.risk)
                        diverse_eff.append(self.getDiversification(self.ww))
                    res['efficient']['efficient'] = len(risk_eff)
                    res['efficient']['risk'] = np.array(risk_eff)
                    res['efficient']['rr'] = np.array(RR_eff)
                    res['efficient']['ww'] = ww_eff
                    res['efficient']['sharpe'] = np.array(sharpe_eff)
                    res['efficient']['diverse'] = np.array(diverse_eff)
    
            # inefficient frontier
            res['inefficient']['inefficient'] = 0
            if inefficient > 0 :
                if rr_min is None:
                    warnings.warn("Warning: MinRisk -> inefficient failed")
                    res['inefficient']['inefficient'] = 0
                else:
                    mu_min = self.muk.min()
                    rr_lower = mu_min + np.abs(mu_min) * _RR_ABS_TOL_
                    rr_ineff = np.linspace(rr_lower, rr_min, inefficient)
                    risk_ineff = []
                    ww_ineff = []
                    RR_ineff = []
                    sharpe_ineff = []
                    diverse_ineff = []
                    for rr in rr_ineff:
                        _ww = self.getWeights(rtype='Risk', mu=rr, d=-1)
                        if self.status != 0:
                            warnings.warn(f"Warning: inefficient for rr {rr} "
                                          "failed")
                            continue
                        ww_ineff.append(_ww)
                        risk_ineff.append(self.risk)
                        RR_ineff.append(self.RR)
                        sharpe_ineff.append(
                            (np.dot(self.ww, self.muk) - musharpe) / self.risk)
                        diverse_ineff.append(self.getDiversification(self.ww))
                    res['inefficient']['inefficient'] = len(risk_ineff)
                    res['inefficient']['risk'] = np.array(risk_ineff)
                    res['inefficient']['rr'] = np.array(RR_ineff)
                    res['inefficient']['ww'] = ww_ineff
                    res['inefficient']['sharpe'] = np.array(sharpe_ineff)
                    res['inefficient']['diverse'] = np.array(diverse_ineff)
    
            # sharpe point
            res['sharpe']['sharpe'] = sharpe
            if sharpe:
                ww_sharpe = self.getWeights(rtype='Sharpe', mu0=musharpe)
                if self.status != 0:
                    res['sharpe']['sharpe'] = False
                    warnings.warn("Warning: MinRisk failed")
                else:
                    rr_sharpe = self.RR
                    risk_sharpe = self.risk
                    sharpe_sharpe = self.sharpe
                    diverse_sharpe = self.getDiversification(ww_sharpe)
                    res['sharpe']['mu'] = musharpe
                    res['sharpe']['risk'] = risk_sharpe
                    res['sharpe']['rr'] = rr_sharpe
                    res['sharpe']['ww'] = ww_sharpe
                    res['sharpe']['sharpe'] = sharpe_sharpe
                    res['sharpe']['diverse'] = diverse_sharpe
                    res['sharpe']['label'] = options['sharpe_label']
    
            # portfolio components
            res['component']['component'] = component
            if component:
                risk_comp = []
                rr_comp = []
                label_comp = []
                ww_comp = []
                sharpe_comp = []
                for k in range(self.mm):
                    ww = [0.] * self.mm
                    ww[k] = 1.
                    risk_comp.append(self.getRisk(ww))
                    if self.status != 0:
                        warnings.warn("Warning: component "
                                      f"{self.rrate.columns[k]} failed")
                        continue
                    rr_comp.append(self.muk[k])
                    label_comp.append(self.muk.index[k])
                    ww_comp.append(ww)
                    sharpe_comp.append(
                        (self.muk[k] - musharpe) / self.risk \
                            if self.risk != 0 else 0)
                if len(risk_comp) == 0:
                    res['component']['component'] = False
                else:
                    res['component']['risk'] = np.array(risk_comp)
                    res['component']['rr'] = np.array(rr_comp)
                    res['component']['ww'] = ww_comp
                    res['component']['label'] = label_comp \
                        if options['comp_label'] else [None] * self.mm
                    res['component']['sharpe'] = np.array(sharpe_comp)
                    res['component']['diverse'] = np.zeros(self.mm)
    
            # random portfolios
            res['randomport']['randomport'] = 0
            if randomport > 0:
                risk_rp = []
                rr_rp = []
                ww_rp = []
                sharpe_rp = []
                diverse_rp = []
                for _ in range(randomport):
                    ww = self._ww_gen()
                    _div = self.getDiversification(ww)
                    if self.status != 0:
                        warnings.warn(f"Warning: randomport {ww} failed")
                        continue
                    diverse_rp.append(_div)
                    risk_rp.append(self.risk)
                    rr_rp.append(self.RR)
                    ww_rp.append(ww)
                    sharpe_rp.append(
                        (np.dot(self.ww, self.muk) - musharpe) / self.risk)
                res['randomport']['randomport'] = len(risk_rp)
                res['randomport']['risk'] = np.array(risk_rp)
                res['randomport']['rr'] = np.array(rr_rp)
                res['randomport']['ww'] = ww_rp
                res['randomport']['sharpe'] = np.array(sharpe_rp)
                res['randomport']['diverse'] = np.array(diverse_rp)
    
            # invN
            res['invN']['invN'] = invN
            if invN:
                res['invN']['ww'] = np.full(self.mm, 1 / self.mm)
                _div = self.getDiversification(res['invN']['ww'])
                if self.status != 0:
                    res['invN']['invN'] = False
                    warnings.warn("Warning: invN failed")
                else:
                    res['invN']['diverse'] = _div
                    res['invN']['risk'] = self.risk
                    res['invN']['rr'] = self.RR
                    res['invN']['label'] = options['invN_label']
                    res['invN']['sharpe'] = \
                        (np.dot(self.ww, self.muk) - musharpe) / self.risk
                
            # invNrisk
            res['invNrisk']['invNrisk'] = invNrisk
            if invNrisk:
                _ww = self.getWeights(rtype="InvNrisk")
                if self.status != 0:
                    res['invNrisk']['invNrisk'] = False
                    warnings.warn("Warning: invNrisk failed")
                else:
                    res['invNrisk']['ww'] = _ww
                    res['invNrisk']['risk'] = self.risk
                    res['invNrisk']['rr'] = self.RR
                    res['invNrisk']['label'] = options['invNrisk_label']
                    res['invNrisk']['sharpe'] = \
                        (np.dot(self.ww, self.muk) - musharpe) / self.risk
                    res['invNrisk']['diverse'] = \
                        self.getDiversification(res['invNrisk']['ww'])
                    
            # invNdiverse
            res['invNdiverse']['invNdiverse'] = invNdiverse
            if invNdiverse:
                _ww = self.getWeights(rtype="InvNdiverse")
                if self.status != 0:
                    res['invNdiverse']['invNdiverse'] = False
                    warnings.warn("Warning: invNdiverse failed")
                else:
                    res['invNdiverse']['ww'] = _ww
                    res['invNdiverse']['risk'] = self.risk
                    res['invNdiverse']['rr'] = self.RR
                    res['invNdiverse']['label'] = options['invNdiverse_label']
                    res['invNdiverse']['sharpe'] = \
                        (np.dot(self.ww, self.muk) - musharpe) / self.risk
                    res['invNdiverse']['diverse'] = self.diverse
                
            # invNdrr
            res['invNdrr']['invNdrr'] = invNdrr
            if invNdrr:
                _ww = self.getWeights(rtype="InvNdrr")
                if self.status != 0:
                    res['invNdrr']['invNdrr'] = False
                    warnings.warn("Warning: maxdiverse failed")
                else:
                    res['invNdrr']['ww'] = _ww
                    res['invNdrr']['risk'] = self.risk
                    res['invNdrr']['rr'] = self.RR
                    res['invNdrr']['label'] = options['invNdrr_label']
                    res['invNdrr']['sharpe'] = \
                        (np.dot(self.ww, self.muk) - musharpe) / self.risk
                    res['invNdrr']['diverse'] = self.diverse
                
            # maxdiverse
            res['maxdiverse']['maxdiverse'] = maxdiverse
            if maxdiverse_calc:
                ww_d = self.getWeights('MaxDiverse')  
                if self.status != 0:
                    res['maxdiverse']['maxdiverse'] = False
                    warnings.warn("Warning: maxdiverse failed")
                else:
                    res['maxdiverse']['risk'] = self.risk 
                    rr_maxdiverse = self.RR 
                    res['maxdiverse']['rr'] = rr_maxdiverse
                    res['maxdiverse']['ww'] = ww_d
                    res['maxdiverse']['diverse'] = self.diverse
                    res['maxdiverse']['label'] = options['maxdiverse_label']
                    res['maxdiverse']['sharpe'] = \
                        (np.dot(ww_d, self.muk) - musharpe) / self.risk
                
            # diverse-efficient frontier
            res['diverse_efficient']['diverse_efficient'] = 0
            if diverse_efficient > 0 :
                if rr_maxdiverse is None:
                    warnings.warn("Warning: maxdiverse -> diverse_efficient "
                                  "failed")
                    res['diverse_efficient']['diverse_efficient'] = 0
                else:
                    mu_max = self.muk.max() 
                    rr_upper = mu_max - np.abs(mu_max) * _RR_ABS_TOL_
                    rr_d_eff = np.linspace(rr_maxdiverse, rr_upper,
                                           diverse_efficient)
                    risk_d_eff = []
                    ww_d_eff = []
                    RR_d_eff = []
                    sharpe_d_eff = []
                    diverse_d_eff = []
                    for rr in rr_d_eff:
                        _ww = self.getWeights(rtype='Diverse', mu=rr)
                        if self.status != 0:
                            warnings.warn("Warning: diverse_efficient "
                                          f"for rr {rr} failed")
                            continue
                        ww_d_eff.append(_ww)
                        risk_d_eff.append(self.risk)
                        RR_d_eff.append(self.RR)
                        sharpe_d_eff.append(
                            (np.dot(self.ww, self.muk) - musharpe) / self.risk \
                                if self.risk != 0 else 0)
                        diverse_d_eff.append(self.getDiversification(self.ww))
                    res['diverse_efficient']['diverse_efficient'] = len(risk_d_eff)
                    res['diverse_efficient']['risk'] = np.array(risk_d_eff)
                    res['diverse_efficient']['rr'] = np.array(RR_d_eff)
                    res['diverse_efficient']['ww'] = ww_d_eff
                    res['diverse_efficient']['sharpe'] = np.array(sharpe_d_eff)
                    res['diverse_efficient']['diverse'] = np.array(diverse_d_eff)
                
            # diverse-inefficient frontier
            res['diverse_inefficient']['diverse_inefficient'] = 0
            if diverse_inefficient > 0 :
                if rr_maxdiverse is None:
                    warnings.warn("Warning: maxdiverse -> diverse_inefficient "
                                  "failed")
                    res['diverse_inefficient']['diverse_inefficient'] = 0
                else:
                    rr_min = self.muk.min()
                    rr_lower = rr_min + np.abs(mu_min) * _RR_ABS_TOL_
                    rr_d_ineff = np.linspace(rr_lower, rr_maxdiverse, 
                                             diverse_inefficient)
                    risk_d_ineff = []
                    ww_d_ineff = []
                    RR_d_ineff = []
                    sharpe_d_ineff = []
                    diverse_d_ineff = []
                    for rr in rr_d_ineff:
                        _ww = self.getWeights(rtype='Diverse', mu=rr, d=-1)
                        if self.status != 0:
                            warnings.warn(f"Warning: diverse_inefficient "
                                          f"for rr {rr} failed")
                            continue
                        ww_d_ineff.append(_ww)
                        risk_d_ineff.append(self.risk)
                        RR_d_ineff.append(self.RR)
                        sharpe_d_ineff.append(
                            (np.dot(self.ww, self.muk) - musharpe) / self.risk \
                                if self.risk != 0 else 0)
                        diverse_d_ineff.append(self.getDiversification(self.ww))
                    res['diverse_inefficient']['diverse_inefficient'] = \
                        len(risk_d_ineff)
                    res['diverse_inefficient']['risk'] = np.array(risk_d_ineff)
                    res['diverse_inefficient']['rr'] = np.array(RR_d_ineff)
                    res['diverse_inefficient']['ww'] = ww_d_ineff
                    res['diverse_inefficient']['sharpe'] = \
                        np.array(sharpe_d_ineff)
                    res['diverse_inefficient']['diverse'] = \
                        np.array(diverse_d_ineff)
                
            # addport
            res['addport']['addport'] = False
            if addport is not None:
                if isinstance(addport, dict):
                    aport = pd.DataFrame().from_dict(addport, 'index', 
                                                    columns=self.rrate.columns)
                elif isinstance(addport, pd.core.frame.DataFrame):
                    aport = addport[self.rrate.columns]
                else:
                    raise ValueError("addport must be a dict or a "
                                     + "pandas.DataFrame, see doc")
     
                def ff_addport(x):
                    diverse_addport = self.getDiversification(x)
                    if self.status != 0:
                        warnings.warn(f"Warning: addport {x} failed")
                        risk_addport = None
                        RR_addport = None
                        sharpe_addport = None
                    else:
                        risk_addport = self.risk
                        RR_addport = self.RR
                        sharpe_addport = (np.dot(x, self.muk) - musharpe) \
                             / risk_addport if risk_addport != 0  else 0 
                             
                    return pd.Series([self.status, risk_addport, RR_addport, 
                        sharpe_addport, diverse_addport],
                        index=['status','risk', 'rr', 'sharpe', 'diverse'])
                
                res_addport = aport.apply(lambda x: ff_addport(x), axis=1).dropna()
                if res_addport.size > 0:
                    res['addport']['addport'] = True
                    res['addport']['risk'] = res_addport['risk'].to_list()
                    res['addport']['rr'] = res_addport['rr'].to_list()
                    res['addport']['sharpe'] = res_addport['sharpe'].to_list()
                    res['addport']['diverse'] = res_addport['diverse'].to_list()
                    res['addport']['ww'] = aport.to_numpy().tolist()
                    res['addport']['label'] = \
                        aport.index.to_list() if options['addport_label'] \
                            else [None] * aport.shape[0]
                    res['addport']['hidden_label'] = aport.index.to_list()
                
            resout = res
            res['options'] = options
        else:
            res = copy.deepcopy(resout)

            res['options'] = options
            res['risk_min']['label'] = options['minrisk_label']
            res['sharpe']['label'] = options['sharpe_label']
            res['component']['label'] = self.muk.index.tolist() \
                if options['comp_label'] else [None] * self.mm
            res['invN']['label'] = options['invN_label']
            res['invNrisk']['label'] = options['invNrisk_label']
            res['invNdiverse']['label'] = options['invNdiverse_label']
            res['invNdrr']['label'] = options['invNdrr_label']
            res['maxdiverse']['label'] = options['maxdiverse_label']
            if res['addport']['addport']:
                if options['addport_label']:
                    res['addport']['label'] = res['addport']['hidden_label']
                else:
                    res['addport']['label'] = [None] * \
                        len(res['addport']['hidden_label'])
            
        # Plot
        plot_ff(res)

        return resout


    def _plot_f1(self, res):
        fig = plt.figure()
        ax = fig.add_subplot(111)
        
        ax.set(title=res['options']['title'], 
               xlabel=res['options']['xlabel'], 
               ylabel=res['options']['ylabel'])

        if res['risk_min']['risk_min']:
            ax.scatter(x=res['risk_min']['risk'], y=res['risk_min']['rr'],
                       marker='D', c='g', s=36)
            ax.annotate(res['risk_min']['label'],
                        (res['risk_min']['risk'] * 1.01,
                         res['risk_min']['rr'] * 1.04))

        if res['efficient']['efficient'] > 0:
            ax.plot(res['efficient']['risk'], res['efficient']['rr'],
                    color='lightblue', linewidth=2)

        if res['inefficient']['inefficient'] > 0:
            ax.plot(res['inefficient']['risk'], res['inefficient']['rr'],
                    color='red', linewidth=2)

        if res['component']['component']:
            lf = 1.01
            for k in range(len(res['component']['risk'])):
                ax.scatter(x=res['component']['risk'][k],
                           y=res['component']['rr'][k],
                           marker='s', c='b', s=16)
                ax.annotate(res['component']['label'][k],
                            (res['component']['risk'][k] * lf,
                             res['component']['rr'][k] * lf))

        if res['randomport']['randomport'] > 0:
            ax.scatter(x=res['randomport']['risk'], y=res['randomport']['rr'],
                       marker='o', s=9, edgecolors='b', alpha=0.5)

        for invstr in ['invN', 'invNrisk', 'invNdiverse', 'invNdrr']:
            if res[invstr][invstr]:
                ax.scatter(x=res[invstr]['risk'], y=res[invstr]['rr'],
                          marker='x', c='g', s=36)
                lf = 1.01
                ax.annotate(res[invstr]['label'],
                            (res[invstr]['risk'] * lf,
                             res[invstr]['rr'] * lf))
  
        if res['maxdiverse']['maxdiverse']:
            ax.scatter(x=res['maxdiverse']['risk'], y=res['maxdiverse']['rr'],
                       marker='x', c='b', s=36)
            lf = 1.01
            ax.annotate(res['maxdiverse']['label'],
                        (res['maxdiverse']['risk'] * lf,
                         res['maxdiverse']['rr'] * lf))
    
        if res['diverse_efficient']['diverse_efficient'] > 0:
            ax.plot(res['diverse_efficient']['risk'], 
                    res['diverse_efficient']['rr'],
                    color='lightblue', linewidth=1, linestyle=':')

        if res['diverse_inefficient']['diverse_inefficient'] > 0:
            ax.plot(res['diverse_inefficient']['risk'], 
                    res['diverse_inefficient']['rr'],
                    color='red', linewidth=1, linestyle=':')
            
        if res['addport']['addport']:
            lf = 1.01
            for k in range(len(res['addport']['risk'])):
                ax.scatter(x=res['addport']['risk'][k],
                           y=res['addport']['rr'][k],
                           marker='o', s=9, edgecolors='r', alpha=0.5)
                ax.annotate(res['addport']['label'][k],
                            (res['addport']['risk'][k] * lf,
                             res['addport']['rr'][k] * lf))

        xl1, xl2 = plt.xlim()
        yl1, yl2 = plt.ylim()
        lff = 1.02
        xl2 *= lff
        yl2 *= lff
        ax.set(xlim=(xl1, xl2), ylim=(yl1, yl2))

        if res['sharpe']['sharpe']:
            ax.scatter(x=res['sharpe']['risk'], y=res['sharpe']['rr'],
                       marker='D', c='g', s=26)
            ax.annotate(res['sharpe']['label'],
                        (res['sharpe']['risk'] * 0.91,
                         res['sharpe']['rr'] * 1.04))
            if res['options']['tangent']:
                sharpe = res['sharpe']['sharpe']
                musharpe = res['sharpe']['mu']
                if np.abs(sharpe) < 1.e-5:
                    ax.axhline(res['sharpe']['rr'])
                else:
                    x1 = xl1
                    y1 = sharpe * x1 + musharpe
                    if y1 < yl1:
                        y1 = yl1
                        x1 = (y1 - musharpe) / sharpe
                    x2 = xl2
                    y2 = sharpe * x2 + musharpe
                    if y2 > yl2:
                        y2 = yl2
                        x2 = (y2 - musharpe) / sharpe
                    ax.plot([x1, x2], [y1, y2], color='black', linewidth=1)
                    
        if res['options']['saveto'] is not None:
            plt.savefig(res['options']['saveto'])
        plt.show()


    def _plot_f2(self, res):
        fig = plt.figure()
        ax = fig.add_subplot(111)
        
        ax.set(title=res['options']['title'], 
               xlabel=res['options']['xlabel'], 
               ylabel=res['options']['ylabel'])

        if res['risk_min']['risk_min']:
            ax.scatter(x=res['risk_min']['rr'], y=res['risk_min']['sharpe'], 
                       marker='D', c='g', s=36)
            ax.annotate(res['risk_min']['label'],
                        (res['risk_min']['rr'] * 0.92,
                         res['risk_min']['sharpe'] * 1.04))

        if res['efficient']['efficient'] > 0:
            ax.plot(res['efficient']['rr'], res['efficient']['sharpe'],
                    color='lightblue', linewidth=2)

        if res['inefficient']['inefficient'] > 0:
            ax.plot(res['inefficient']['rr'], res['inefficient']['sharpe'],
                    color='red', linewidth=2)

        if res['component']['component']:
            lf = 1.01
            for k in range(len(res['component']['risk'])):
                ax.scatter(x=res['component']['rr'][k], 
                           y=res['component']['sharpe'][k],
                           marker='s', c='b', s=16)
                ax.annotate(res['component']['label'][k],
                            (res['component']['rr'][k] * lf, 
                             res['component']['sharpe'][k] * lf))
            
        if res['sharpe']['sharpe']:
            ax.scatter(x=res['sharpe']['rr'], y=res['sharpe']['sharpe'],
                       marker='D', c='g', s=26)
            ax.annotate(res['sharpe']['label'],
                        (res['sharpe']['rr'] * 0.93,
                         res['sharpe']['sharpe'] * 1.02))

        if res['randomport']['randomport'] > 0:
            ax.scatter(x=res['randomport']['rr'], 
                       y=res['randomport']['sharpe'],
                       marker='o', s=9, edgecolors='b', alpha=0.5 )

        for invstr in ['invN', 'invNrisk', 'invNdiverse', 'invNdrr']:
            if res[invstr][invstr]:
                ax.scatter(x=res[invstr]['rr'], y=res[invstr]['sharpe'],
                          marker='x', c='g', s=36)
                lf = 1.01
                ax.annotate(res[invstr]['label'],
                            (res[invstr]['rr'] * lf,
                             res[invstr]['sharpe'] * lf))
   
        if res['maxdiverse']['maxdiverse']:
            ax.scatter(x=res['maxdiverse']['rr'], 
                       y=res['maxdiverse']['sharpe'],
                       marker='x', c='b', s=36)
            lf = 1.01
            ax.annotate(res['maxdiverse']['label'],
                       (res['maxdiverse']['rr'] * lf, 
                        res['maxdiverse']['sharpe'] * lf))

        if res['diverse_efficient']['diverse_efficient'] > 0:
            ax.plot(res['diverse_efficient']['rr'], 
                    res['diverse_efficient']['sharpe'],
                    color='lightblue', linewidth=1, linestyle=':')

        if res['diverse_inefficient']['diverse_inefficient'] > 0:
            ax.plot(res['diverse_inefficient']['rr'], 
                    res['diverse_inefficient']['sharpe'],
                    color='red', linewidth=1, linestyle=':')
            
        if res['addport']['addport']:
            lf = 1.01
            for k in range(len(res['addport']['risk'])):
                ax.scatter(x=res['addport']['rr'][k],
                           y=res['addport']['sharpe'][k],
                           marker='o', s=9, edgecolors='r', alpha=0.5)
                ax.annotate(res['addport']['label'][k],
                            (res['addport']['rr'][k] * lf,
                             res['addport']['sharpe'][k] * lf))
            
        xl1, xl2 = plt.xlim()
        yl1, yl2 = plt.ylim()
        lff = 1.02
        xl2 *= lff
        yl2 *= lff
        ax.set(xlim=(xl1, xl2), ylim=(yl1, yl2))

        if res['options']['saveto'] is not None:
            plt.savefig(res['options']['saveto'])
        plt.show()


    def _plot_f3(self, res):
        fig = plt.figure()
        ax = fig.add_subplot(111)

        ax.set(title=res['options']['title'], 
               xlabel=res['options']['xlabel'], 
               ylabel=res['options']['ylabel'])

        if res['risk_min']['risk_min']:
            ax.scatter(x=res['risk_min']['rr'], y=res['risk_min']['diverse'], 
                       marker='D', c='g', s=36)
            lf = 0.95
            ax.annotate(res['risk_min']['label'],
                       (res['risk_min']['rr'] * lf, 
                        res['risk_min']['diverse'] * lf))
    
        if res['efficient']['efficient'] > 0:
            ax.plot(res['efficient']['rr'], res['efficient']['diverse'],
                    color='lightblue', linewidth=1, linestyle=':')
    
        if res['inefficient']['inefficient'] > 0:
            ax.plot(res['inefficient']['rr'], res['inefficient']['diverse'],
                    color='red', linewidth=1, linestyle=':')
    
        if res['component']['component']:
            lf = 1.01
            for k in range(len(res['component']['risk'])):
                ax.scatter(x=res['component']['rr'][k], 
                           y=res['component']['diverse'][k],
                           marker='s', c='b', s=16)
                ax.annotate(res['component']['label'][k],
                            (res['component']['rr'][k] * lf, 
                             res['component']['diverse'][k] * lf))
    
        if res['sharpe']['sharpe']:
            ax.scatter(x=res['sharpe']['rr'], 
                       y=res['sharpe']['diverse'],
                       marker='D', c='g', s=26)
            lf = 1.01
            ax.annotate(res['sharpe']['label'],
                       (res['sharpe']['rr'] * lf, 
                        res['sharpe']['diverse'] * lf))
    
        if res['randomport']['randomport'] > 0:
            ax.scatter(x=res['randomport']['rr'], 
                       y=res['randomport']['diverse'],
                       marker='o', s=9, edgecolors='b', alpha=0.5 )
    
        for invstr in ['invN', 'invNrisk', 'invNdiverse', 'invNdrr']:
            if res[invstr][invstr]:
                ax.scatter(x=res[invstr]['rr'], 
                           y=res[invstr]['diverse'],
                           marker='x', c='g', s=36)
                lf = 1.01
                ax.annotate(res[invstr]['label'],
                            (res[invstr]['rr'] * lf,
                             res[invstr]['diverse'] * lf))
 
        if res['maxdiverse']['maxdiverse']:
            ax.scatter(x=res['maxdiverse']['rr'], 
                       y=res['maxdiverse']['diverse'],
                       marker='x', c='b', s=36)
            lf = 1.01
            ax.annotate(res['maxdiverse']['label'],
                       (res['maxdiverse']['rr'] * lf, 
                        res['maxdiverse']['diverse'] * lf))
            
        if res['diverse_efficient']['diverse_efficient'] > 0:
            ax.plot(res['diverse_efficient']['rr'], 
                    res['diverse_efficient']['diverse'],
                    color='lightblue', linewidth=2)
    
        if res['diverse_inefficient']['diverse_inefficient'] > 0:
            ax.plot(res['diverse_inefficient']['rr'], 
                    res['diverse_inefficient']['diverse'],
                    color='red', linewidth=2)
            
        if res['addport']['addport']:
            lf = 1.01
            for k in range(len(res['addport']['risk'])):
                ax.scatter(x=res['addport']['rr'][k],
                           y=res['addport']['diverse'][k],
                           marker='o', s=9, edgecolors='r', alpha=0.5)
                ax.annotate(res['addport']['label'][k],
                            (res['addport']['rr'][k] * lf,
                             res['addport']['diverse'][k] * lf))
            
        xl1, xl2 = plt.xlim()
        yl1, yl2 = plt.ylim()
        lff = 1.02
        xl2 *= lff
        yl2 *= lff
        ax.set(xlim=(xl1, xl2), ylim=(yl1, yl2))
    
        if res['options']['saveto'] is not None:
            plt.savefig(res['options']['saveto'])
        plt.show()


    def set_random_seed(self, seed = 42):
        """
        Sets the seed for Dirichlet random generator used in viewFrontiers
        to select the random inefficient portfolios.

        Parameters
        ----------
        `seed` : `int`, optional;
            The random generator seed, in case you want to set it to a weird
            value other than `42` :). The default is `42`.

        Returns
        -------
        `None`
        """
        self.rng = np.random.RandomState(seed)


    def _ww_gen(self):
        return self.rng.dirichlet([0.5] * self.mm)
    
    
    def _set_lp_method(self, method):
        self.methods = ['ecos', 'highs-ds', 'highs-ipm', 'highs',
                       'interior-point', 'glpk', 'cvxopt']
        if not method in self.methods:
            raise ValueError(f"unknown method {method} - "
                             f"must be one of {self.methods}")
        self.method = method
        
        
    def _set_qp_method(self, method):
        self.methods = ['ecos', 'cvxopt']
        if not method in self.methods:
            raise ValueError(f"unknown method {method} - "
                             f"must be one of {self.methods}")
        self.method = method
        
        
    def _set_socp_method(self, method):
        self.methods = ['ecos', 'cvxopt']
        if not method in self.methods:
            raise ValueError(f"unknown method {method} - "
                             f"must be one of {self.methods}")
        self.method = method
        

    
    def getRiskComp(self):
        """
        Returns the risk of each portfolio component.

        Returns
        -------
        `pandas.Series`
        """
        if self._flag_risk_comp_calc: 
            self.status = self._status_risk_comp_calc
            return self.risk_comp
            
        srisk = np.zeros(self.rrate.shape[1])
        status = 0
        for i in range(len(srisk)):
            wws = np.zeros(len(srisk))
            wws[i] = 1
            srisk[i] = self.getRisk(wws)
            if self.status != 0:
                status = self.status
                srisk[i] = None
 
        self.risk_comp = pd.Series(srisk, self.rrate.columns)
        self._flag_risk_comp_calc = True
        self.status = status
        self._status_risk_comp_calc = self.status
        
        if self.status != 0:
            warnings.warn(f"Fail to compute for edate {self.rrate.index[-1]} "
                          f"all risk coml\n{self.risk_comp}")

        return self.risk_comp
        
    
    def getDiversification(self, ww, rrate=None):
        """
        Returns the value of the diversification factor for a give 
        portfolio.

        Parameters
        ----------
        `ww` : `list`, `numpy.array` or `pandas.Series`;
            Portfolio weights. Its length must be equal to the number of
            symbols in `rrate` (mktdata). All weights must be >=0 with 
            sum > 0.
            If it is a list or a `numpy.array` then the weights are assumed to
            be in order of `rrate.columns`. If it is a `pandas.Series` then 
            the index should be compatible with the `rrate.columns` or mktdata 
            symbols (not necessary in the same order).
        `rrate` : `pandas.DataFrame`, optional;
            Contains the portfolio components historical
            rates of returns. If it is not `None` then it will overwrite the
            `rrate` computed in the constructor from `mktdata`.
            The default is `None`.

        Returns
        -------
        `float` : The diversification value. \n
        """
        self._reset_output()
        
        if rrate is not None:
            self.set_rrate(rrate)

        if isinstance(ww, pd.core.series.Series):
            ww = ww[self.rrate.columns]
        w = np.array(ww)
        if any(w < 0.):
            raise ValueError("All ww must be non-negative")
        sw = w.sum()
        if sw <= 0:
            raise ValueError("At least one ww must be positive")
        w = w / sw
        
        self.getRiskComp()
        if self.status != 0:
            warnings.warn("Warning: Fail to compute all risk comp\n"
                          f"{self.risk_comp}")
            return None
            
        wrisk = self.getRisk(w)
        if self.status !=0:
            warnings.warn(f"Warning: Fail to compute risk for {ww}")
            return None
        
        crisk = np.dot(self.risk_comp, w)

        if crisk == 0:
            self.diverse = 0
        else:
            self.diverse = 1 - wrisk / crisk
        
        return self.diverse
    
    
    def set_method(self, method):
        """
        Set computation method

        Parameters
        ----------
        `method` : `str`;
            Must be a valid method name. It will overright the value 
            set by the constructor.

        Returns
        -------
        `None`
        """
        self._set_method(method)
    
    
    def _norm_weights(self):
        self.ww[self.ww < _WW_TOL_] = 0.
        self.ww /= self.ww.sum()
        
        
    def _single_asset_port(self, port):
        if port == 'min':
            psymb = self.muk.idxmin()
        elif port == 'max':
            psymb = self.muk.idxmax()
        elif port in self.rrate.columns:
            psymb = port
        else:
            raise ValueError(f"argumkent ({port}) must be "
                             f"one of 'min', 'max', {self.symb}")
        w = pd.Series(0., index=self.rrate.columns)
        w[psymb] = 1.
        self.getRisk(w) 
        self.diverse = 0.


    # to be implemented in the deriv classes
    def _risk_calc(self, prate, alpha):
        # Computes a simple risk measure for `prate` time series and `alpha`
        # parameter (where is the case)
        pass
    def _risk_min(self, d=1):
        # Computes the optimal portfolio with minimum risk for fixed expected 
        # rate of return where `d` define the risk frontier type: 
        # +1 for efficient and -1 for inefficient 
        pass
    def _sharpe_max(self):
        # Computes the optimal portfolio with maximum Sharpe ratio
        pass
    def _sharpe_inv_min(self):
        # Computes the optimal portfolio with minimum inverse Sharpe ratio
        pass
    def _rr_max(self):
        # Computes the optimal-risk portfolio with maximum expected rate of 
        # return for fixed risk
        pass
    def _risk_averse(self):
        # Computes the optimal-risk portfolio for fixed risk-aversion factor
        pass
    def _risk_diversification(self, d=1):
        # Computes the optimal-diversified portfolio for fixed expected
        # rate of return where `d` define the diversified frontier type: 
        # +1 for efficient and -1 for inefficient.
        # Minimum of 1-D.
        pass
    def _risk_inv_diversification(self, d=1):
        # Computes the optimal diversified portfolio for fixed expected
        # rate of return where `d` define the diversified frontier type: 
        # +1 for efficient and -1 for inefficient.
        # Maximum of inverse 1-D. 
        pass
    def _rr_max_diversification(self):
        # Computes the optimal diversified portfolio for fixed diversification
        # factor. Maximization of expected rate of return.
        pass
    def _set_method(self, method):
        # sets method
        pass
    


"""
Contains:
    
    - schedule_simple : creates a simple schedule
    - schedule_roll : creates a rollover schedule
"""

import pandas as pd
import numpy as np
import pandas.tseries.offsets as pt

from azapy.MkT.MkTcalendar import NYSEgen

def schedule_simple(sdate='2010-01-01',
                    edate='today',
                    freq='Q',
                    noffset=-3,
                    fixoffset=-1,
                    calendar=None):
    """
    Creates a simple schedule 'Droll', 'Dfix'.

    Parameters
    ----------
    `sdate` : `str`, optional;
        Start date (reference) of the schedule.
        The default is `'2010-01-01'`.
    `edate` : `str`, optional;
        End date (reference) of the schedule. 
        The default is `'today'`.
    `freq` : `str`, optional;
        Rolling period. It can take 2 values: `'Q'` for quarterly and `'M'` for
        monthly rolling periods. The default is `'Q'`.
    `noffset` : `int`, optional;
        Offset in number of business days for `Droll` relative to the end of 
        calendar period (quarter or month). The default is `-3`.
    `fixoffset` : `int`, optional
        Offset in number of business days for `Dfix` relative to `Droll`. It 
        must be <=0. The default is `-1`.
    `calendar` : `numpy.busdaycalendar`, optional;
        Business days calendar. If is it `None` then the calendar will be set
        to NYSE business calendar.
        The default is `None`.

    Returns
    -------
    `pandas.DataFrame`
        Table containing 2 datetime columns: 'Droll' the rolling date and
        'Dfix' the fixing date.
    """
    sdate = pd.to_datetime(sdate)
    if freq == 'Q': edate = pd.to_datetime(edate) + pt.QuarterEnd(1)
    elif freq == 'M': edate = pd.to_datetime(edate) + pt.MonthEnd(1)
    else: raise ValueError("Wrong freq, Must be 'Q' or 'M'")
    
    if calendar is None:
        calendar = NYSEgen()
    
    tedx = pd.date_range(start=sdate, end=edate, freq=freq)\
        .to_numpy(dtype='<M8[D]')
    troll = np.busday_offset(tedx, noffset, roll='backward', 
                             busdaycal=calendar)
    tfix = np.busday_offset(troll, fixoffset, roll='backward', 
                            busdaycal=calendar)
    return pd.DataFrame({'Droll': troll, 'Dfix': tfix})

def schedule_roll(sdate='2010-01-01',
                  edate='today',
                  freq='Q',
                  noffset=-3,
                  fixoffset=-1,
                  calendar=None,
                  hlength=1.25):
    """
    Creates a schedule with rolling history: 'Droll', 'Dfix' and 'Dhist'.

    Parameters
    ----------
    `sdate` : `str`, optional;
        Start date (reference) of the schedule.
        The default is `'2010-01-01'`.
    `edate` : `str`, optional;
        End date (reference) of the schedule. 
        The default is `'today'`.
    `freq` : `str`, optional;
        Rolling period. It can take 2 values: `'Q'` for quarterly and `'M'` for
        monthly rolling periods. The default is `'Q'`.
    `noffset` : `int`, optional;
        Offset in number of business days for Droll relative to the end of 
        calendar period (quarter or month). The default is `-3`.
    `fixoffset` : `int`, optional;
        Offset in number of business days for `Dfix` relative to `Droll`. It 
        can be zero or negative. The default is `-1`.
    `calendar` : `numpy.busdaycalendar`, optional;
        Business days calendar. If is it `None` then the calendar will be set
        to NYSE business calendar.
        The default is `None`.
    `hlength` : `float`, optional;
        Offset in number of years for 'Dhist' relative to 'Dfix'. A fractional 
        value will be rounded to an integer number of months via
        `round(hlength * 12, 0)`. `hlength` must be non-negative.
        The default is `1.25` years.

    Returns
    -------
    `pandas.DataFrame`;
        Table containing 2 datetime columns: 'Droll' the rolling date,
        'Dfix' the fixing date and 'Dhist' start day for a calibration period.
    """
    sdate = pd.to_datetime(sdate)
    edate = pd.to_datetime(edate)
    if calendar is None:
        calendar = NYSEgen()
    sch = schedule_simple(sdate, edate, freq, noffset, fixoffset, calendar)
    sch['Dhist'] = sch['Dfix'] - pd.offsets \
        .DateOffset(months=round(hlength * 12, 0))
    sch['Dhist'] = np.busday_offset(sch['Dhist'].to_numpy(dtype='<M8[D]'), 
                                    0, roll='backward', 
                                    busdaycal=calendar)
    for k in range(len(sch)):
        if sch['Dhist'][k] >= sdate: 
            schedule = sch[k:].reset_index(drop=True)
            return schedule
    
    raise ValueError("Cannot make a schedule!!")
    


import pandas as pd
import numpy as np
from collections import defaultdict 

#from .readMkTData import NYSEgen
from .MkTcalendar import NYSEgen

def summary_MkTData(mktdata, calendar=None, sdate=None, edate=None):
    """
    Summary of MkT data time-series length and quality (checks for missing
    records).

    Parameters
    ----------
    `mktdata` :`pandas.DataFrame` or a dict of `pndas.DataFrame';
        Market Data in the format returned by `azapy.readMkT` function.
    `calendar` : `numpy.busdaycalendar`, optional;
        Business days calendar. If is set to None it will 
        default to NYSE business calendar.
    `sdate` : `pandas.Timestamp`, optional;
        Time-series start date. If it is `None` then `sdate` will be set to the 
        earliest date in mktdata.
        The default is `None`.
    `edate` : `pandas.Timestamp`, optional;
        Time-series end date. If it is `None` then edate will be set to 
        the most recent date in mktdata.
        The default is `None`.

    Returns
    -------
    `pandas.DataFrame`; A table with columns:
        - `symbol` : time-series symbol
        - `begin` : start date
        - `end` : end date
        - `length` : number of records
        - `na_total` : total number of `nan`
        - `na_b` : number of missing records at the beginning
        - `na_e` : number of missing records at the end
        - `cont` : total number of missing records
            
    Comment: the main application is to assess the missing data in the 
    time-series extracted with `azapy.readMkT` function.
    """
    if isinstance(mktdata, dict):
        gite = mktdata.items()
    else:
        gite = mktdata.groupby('symbol')
    
    sds = []
    eds = []
    for _ , v in gite:
        sds.append(v.index[0])
        eds.append(v.index[-1])
    msds = min(sds)
    meds = max(eds)
    if (sdate is None) or (sdate > msds): sdate = msds
    if (edate is None) or (edate < meds): edate = meds
    
    if calendar is None: 
        calendar = NYSEgen()

    hd = pd.DatetimeIndex([dd  for dd in pd.date_range(sdate, edate) \
                           if np.is_busday(dd.date(), busdaycal=calendar)])
    
    res = defaultdict(lambda: [])
    for k, v in gite:
        sd = v.index[0]
        ed = v.index[-1]
        res['symbol'].append(k)
        res['begin'].append(sd)
        res['end'].append(ed)
        res['length'].append(len(v))
        res['na_total'].append(v.isnull().sum().sum())
        res['na_b'].append(0 if sd == hd[0] else hd.get_loc(sd) )
        res['na_e'].append(0 if ed == hd[-1] else len(hd) - hd.get_loc(ed) - 1)
        res['cont'].append(len(hd) - len(v.index))

    return pd.DataFrame(res)



import pandas as pd
import numpy as np
import ta
import numbers
import plotly.graph_objects as go
from collections import defaultdict

from azapy.util.drawdown import drawdown, max_drawdown

def _color_negative_red(val):
    if isinstance(val, numbers.Number) and (val < 0):
        return 'color: red'
    else:
        return 'color: black'


class Port_Simple:
    """
    Back testing the Buy and Hold portfolio.

    Methods:
        * set_model
        * get_port
        * get_mktdata
        * port_view
        * port_view_all
        * port_drawdown
        * port_perf
        * port_annual_returns
        * port_monthly_returns
    """

    def __init__(self, mktdata, symb=None, sdate=None, edate=None,
                 col='adjusted', pname='Port', pcolname=None,
                 capital=100000):
        """
        Constructor

        Parameters
        ----------
        `mktdata` : `pandas.DataFrame`;
            MkT data in the format "symbol", "date", "open", "high", "low",
            "close", "volume", "adjusted", "divd", "split" (e.g. as returned
            by `azapy.readMkT` function).
        `symb` : `list`, optional;
            List of symbols for the basket components. All symbols MkT data
            should be included in `mktdata`. If set to `None` the `symb` will 
            be  set to the full set of symbols included in `mktdata`. 
            The default is `None`.
        `sdate` : date like, optional;
            Start date for historical data. If set to `None` the `sdate` will
            be set to the earliest date in `mktdata`. The default is `None`.
        `edate` : date like, optional;
            End date for historical dates and so the simulation. Must be
            greater than `sdate`. If it is `None` then `edate` will be set
            to the latest date in `mktdata`. The default is `None`.
        `col` : `str`, optional;
            Name of column in the mktdata DataFrame that will be considered
            for portfolio aggregation. The default is 'adjusted'.
        `pname` : `str`, optional;
            The name of the portfolio. The default is 'Port'.
        `pcolname` : `str`, optional;
            Name of the portfolio price column. If it set to `None` that
            `pcolname=pname`. The default is `None`.
        `capital` : `float`, optional;
            Initial portfolio Capital in dollars. The default is 100000.

        Returns
        -------
        The object.
        """
        if isinstance(mktdata, list):
            mktdata = pd.concat(mktdata, axis=1, join='inner') \
                .melt(var_name='symbol', value_name=col, ignore_index=False)

        # set sdate
        mkt_sdate = mktdata.groupby('symbol') \
                           .apply(lambda x: x.index[0]).max()
        if sdate is None:
            self.sdate = mkt_sdate
        else:
            v_date = pd.to_datetime(sdate).normalize().tz_localize(None)
            if v_date < mkt_sdate:
                self.sdate = mkt_sdate
            else:
                self.sdate = v_date
        # set edate
        mkt_edate = mktdata.groupby('symbol') \
                           .apply(lambda x: x.index[-1]).min()
        if edate is None:
            self.edate = mkt_edate
        else:
            v_date = pd.to_datetime(edate).normalize().tz_localize(None)
            if v_date > mkt_edate:
                self.edate = mkt_edate
            else:
                self.edate = v_date
        # set symb
        mkt_symb = mktdata.symbol.unique()
        if symb is None:
            self.symb = mkt_symb
        else:
            if not all(sy in mkt_symb for sy in symb):
                raise ValueError(f"symb list {symb} incompatible"
                                 f" with mktdata {mkt_symb}")

            self.symb = pd.Series(symb)

        # set mktdata
        self.mktdata = mktdata[(mktdata.index >= self.sdate) &
                               (mktdata.index <= self.edate) &
                               mktdata.symbol.isin(self.symb)].copy()

        self.col = col
        self.pname = pname
        self.pcolname = self.pname if pcolname is None else pcolname
        self.capital = capital

        self.ww = None
        self.port = None
        
        self.options = None

    def set_model(self, ww=None):
        """
        Set model parameters and evaluate the portfolio time-series.

        Parameters
        ----------
        `ww` : `list` (also `numpy.array` or `pandas.Series`), optional;
            List of weights. If it is `pandas.Series` the index should match
            the basket `symb`. Otherwise the weights are considered in the 
            `symb` order. If it is set to `None` than `ww` will be set 
            to equal weights.
            The default is `None`.

        Returns
        -------
        pandas.DataFrame
            The portfolio time-series in the format "date", "pcolname".
        """
        # Validate the weights
        if ww is None:
            self.ww = pd.Series(1., index=self.symb)
        elif isinstance(ww, pd.core.series.Series):
            self.ww = ww
        else:
            self.ww = pd.Series(ww, index=self.symb)

        if self.ww.size != self.symb.size:
            raise ValueError(f"wrong ww size - must be {self.symb.size}")
        if np.any(self.ww < 0.):
            raise ValueError("ww elements must be >= 0")
        wws = self.ww.sum()
        if wws <= 0.:
            raise ValueError("At least one ww element must be > 0")
        self.ww = self.ww / wws

        # Calculate
        self._port_calc()
        return self.port

    def _port_calc(self):
        mktdata = self.mktdata.pivot(columns='symbol', values=self.col)
        divid = self.ww / mktdata.iloc[0] * self.capital

        self.port = pd.DataFrame(mktdata @ divid, columns = [self.pcolname])

    def get_port(self):
        """
        Returns the portfolio time-series.

        Returns
        -------
        `pandas.DataFrame`
        """
        return self.port.copy()

    def get_mktdata(self):
        """
        Returns the actual market data used for portfolio evaluations.

        Returns
        -------
        `pandas.DataFrame`
        """
        return self.mktdata.copy()

    def port_view(self, emas=[30, 200], bollinger=False, **opt):
        """
        Plot the portfolio time series together with optional technical
        indicators.

        Parameters
        ----------
        `emas` : `list` of int, optional;
            List of EMA durations. The default is [30, 200].
        `bollinger` : Boolean, optional;
            If set `True` it adds the Bollinger bands. The default is `False`.    
        `opt` : other parameters:
            * `fancy` : Boolean, optional;
                - `False` : it uses the matplotlib capabilities.
                - `True` : it uses plotly library for interactive time-series view.
    
                The default is `False`.
            * `title` : `str`, optional; plot title. The default is `'Port performance'`.
            * `xlabel` : `str`, optional; name of x-axis. The default is `'date'`.
            * `ylabel` : `srt`; optional; name of y-axis. The default is `None`.
            * `saveto` : `str`, optional
                The name of the file where to save the plot. The default is `None`.
        Returns
        -------
        `pandas.DataFrame`;
            Contains the time-series included in plot.
        """
        options = defaultdict(lambda: None, 
                              {'title': "Port performance",
                               'xlabel': "date", 
                               'fancy': False})
        options.update(opt)
        fancy = options['fancy']
        saveto = options['saveto']

        df = self.port.copy()

        if isinstance(emas, list):
            for ema in emas:
                name = 'EMA' + str(ema)
                idx = ta.trend.EMAIndicator(df[self.pcolname], window=ema,
                                            fillna=True)
                df[name] = idx.ema_indicator()

        if bollinger:
            idx = ta.volatility.BollingerBands(df[self.pcolname], fillna=True)
            df['B_upper'] = idx.bollinger_hband()
            df['B_mvag'] = idx.bollinger_mavg()
            df['B_lower'] = idx.bollinger_lband()

        if fancy:
            fig = self._view_plotly(df, options)
            if saveto is not None:
                fig.write_image(saveto)
        else:
            if saveto is None:
                df.plot(title=options['title'], 
                        xlabel=options['xlabel'], 
                        ylabel=options['ylabel'])
            else:
                df.plot(title=options['title'], 
                        xlabel=options['xlabel'], 
                        ylabel=options['ylabel']).get_figure().savefig(saveto)
                        
        return df

    def port_view_all(self, sdate=None, edate=None, componly=False, **opt):
        """
        Plot the portfolio and its component time-series in a relative bases.

        Parameters
        ----------
        `sdate` : date like, optional;
            Start date of plotted time-series. If it is set to `None`
            then the `sdate` is set to the earliest date in the time-series.
            The default is `None`.
        `edate` : date like, optional;
            End date of plotted time-series. If it set to `None` then 
            the `edate
            is set to the most recent date of the time-series.
            The default is `None`.
        `componly` : Boolean, optional;
            `True` : only the portfolio components time-series are plotted.

            `False`: the portfolio and its components times-series are plotted.

            The default is `True`.
        `opt` : other parameters:
            * `fancy` : Boolean, optional;
                - `False` : it uses the pandas plot (matplotlib) capabilities.
                - `True` : it uses plotly library for interactive time-series view.

                The default is `False`.
            * `title` : `str`, optional; plot title. The default is `'Relative performance'`.
            * `xlabel` : `str`, optional; name of x-axis. The default is `'date'`.
            * `ylabel` : `srt`; optional; name of y-axis. The default is `None`.
            * `saveto` : `str`, optional;
                The name of the file where to save the plot. The default is `None`.

        Returns
        -------
        `pandas.DataFrame`;
            A Data Frame containing the time-series.
        """
        options = defaultdict(lambda: None, 
                              {'title': "Relative perfromance",
                               'xlable': "date", 
                               'fancy': False})
        options.update(opt)
        fancy = options['fancy']
        saveto = options['saveto']
        
        
        if sdate is None:
            sdate = self.sdate
        if edate is None:
            edate = self.edate
            
        sdate = pd.to_datetime(sdate)
        edate = pd.to_datetime(edate)

        df = self.mktdata.pivot(columns='symbol', values=self.col)
        df = df.iloc[(df.index >= sdate) & (df.index <= edate)]
        if not componly:
            df = df.merge(self.port, on='date')
        df = df.apply(lambda x: x / x[0])

        if fancy:
            fig = self._view_plotly(df, options)
            if saveto is not None:
                fig.write_image(saveto)
        else:
            if saveto is None:
                df.plot(title=options['title'], 
                        xlabel=options['xlabel'], 
                        ylabel=options['ylabel'])
            else:
                df.plot(title=options['title'], 
                        xlabel=options['xlabel'], 
                        ylabel=options['ylabel']).get_figure().savefig(saveto)

        return df

    def port_drawdown(self, top=5, fancy=False):
        """
        Computes the portfolio drawdowns.

        Parameters
        ----------
        `top` : `int`, optional;
            The number of largest drawdowns that will be reported.
            The default is `5`.
        `fancy` : Boolean, optional;
            `False` : The drawdowns values are reported in unaltered
            algebraic format.

            `True` : The drawdowns values are reported in percent
            rounded to 2 decimals.

            The default is `False`.

        Returns
        -------
        `panda.DataFrame`;
            Table of drawdown events. Columns: \n
                `'DD'` : drawdown rate \n
                `'Date'` : recorded date of the drawdown \n
                `'Star'` : start date of the drawdown \n
                `'End'` : end date of the drawdown
        """
        res = drawdown(self.port, col=self.pcolname, top=top)
        if not fancy: return res

        res.DD = res.DD.round(4) * 100
        return res

    def port_perf(self, componly=False, fancy=False):
        """
        Brief description of portfolio and its components performances
        in terms of average historical rate of returns and maximum drawdowns.

        Parameters
        ----------
        `componly` : Boolean, optional;
            If `True`, only the portfolio components maximum drawdowns
            are reported. The default is `False`.
        `fancy` : Boolean, optional;
            * `False` : The rate of returns and drawdown values are reported
            in unaltered algebraic format.

            * `True` : The rate of returns and drawdown values are reported
            in  percent rounded to 2 decimals.

            The default is `False`.

        Returns
        -------
        `pandas.DataFrame`;
            Performance information.
            Columns:

            * `'RR'` : rate of returns
            * `'DD'` : maximum rate of drawdown
            * `'Beta'` : abs(RR/DD)
            * `'DD_date'` : recorder date of maximum drawdown
            * `'DD_start'` : start date of maximum drawdown
            * `'DD_end'` : end date of maximum drawdown
        """
        # local function
        def rinfo(df, col):
            rr = (df[col][-1] / df[col][0]) ** (252. / len(df)) - 1
            dv, dd, ds, de = max_drawdown(df, col=col)

            return pd.DataFrame([[rr, dv, np.abs(rr / dv), dd, ds, de]],
                columns=['RR', 'DD', 'Beta', 'DD_date', 'DD_start', 'DD_end' ])

        res = self.mktdata.groupby('symbol') \
            .apply(rinfo, col=self.col) \
            .droplevel(1) \
            .sort_values(by='Beta', ascending=False)

        if not componly:
            res2 = rinfo(self.port, self.pcolname)
            res2.index = pd.Index([self.pname], name='symbol')
            res = pd.concat([res2, res])

        if not fancy:
            return res

        res.RR = res.RR.round(4) * 100
        res.DD = res.DD.round(4) * 100
        return res

    def port_annual_returns(self, withcomp=False, componly=False, fancy=False):
        """
        Portfolio annual (calendar) rates of returns.

        Parameters
        ----------
        `withcomp` : Boolean, optional;
            If `True`, adds the portfolio components annual returns to the
            report. The default is `False`.
        `componly` : Boolean, optional;
            If `True`, only the portfolio components annual returns are 
            reported.
            The flag is active only if `withcomp=True`. The default is `False`.
        `fancy` : Boolean, optional;
            `False` : The rates are reported in unaltered
            algebraic format.

            `True` :The rates are reported in percent rounded to 2 decimals
            and presented is color style.

            The default is `False`.

        Returns
        -------
        `pandas.DataFrame`
        """
        # local function
        def frrate(df):
            return df[-1] / df[0] - 1

        if withcomp:
            zz = self.mktdata.pivot(columns='symbol', values=self.col) \
                .loc[self.port.index]
            if not componly:
                zz = self.port.merge(zz, on='date')
            res = zz.resample('A', convention='end').apply(frrate)
        else:
            res = self.port.resample('A', convention='end').apply(frrate)
        res.index = res.index.year
        res.index.name = 'year'

        if not fancy:
            return res

        return res.style.format("{:.2%}").applymap(_color_negative_red)
        

    def port_monthly_returns(self, withcomp=False, componly=False,
                             fancy=False):
        """
        Portfolio monthly (calendar) rate of returns.

        Parameters
        ----------
        `withcomp` : Boolean, optional;
            If `True`, adds the portfolio components monthly returns to the
            report. The default is `False`.
        `componly` : Boolean, optional;
            If `True`, only the portfolio components monthly returns are
            reported. The flag is active only if `withcomp=True`.
            The default is `False`.
        `fancy` : Boolean, optional;
            `False` : The rates are reported in unaltered
            algebraic format.

            `True` :The rates are reported in percent rounded to 2 decimals
            and presented is color style.

            The default is `False`.

        Returns
        -------
        `pandas.DataFrame`
        """
        def frrate(df):
            return df[-1] / df[0] - 1

        # res = self.port.resample('M', convention='end').apply(frrate)
        if withcomp:
            zz = self.mktdata.pivot(columns='symbol', values=self.col) \
                .loc[self.port.index]
            if not componly:
                zz = self.port.merge(zz, on='date')
            res = zz.resample('M', convention='end').apply(frrate)
        else:
            res = self.port.resample('M', convention='end').apply(frrate)

        if not fancy:
            return res

        res.index = pd.MultiIndex.from_arrays([res.index.year.to_list(),
                                               res.index.month.to_list()],
                                              names= ['year', 'month'])

        if not withcomp:
            res = res.reset_index(level=0).pivot(columns='year',
                                                 values=self.pcolname).round(4)

        return res.style.format("{:.2%}").applymap(_color_negative_red)


    def _view_plotly(self, df, options):
        # Create figure
        fig = go.Figure()

        for col in df.columns:
            fig.add_trace(go.Scatter(x=list(df.index), y=list(df[col]),
                                     name=col, line=dict(width=1.5)))

        # Set title
        fig.update_layout(
            title_text=options['title'],
            xaxis_title=options['xlabel'],
            yaxis_title=options['ylabel']
        )

        # Add range slider
        fig.update_layout(
            xaxis=dict(
                rangeselector=dict(
                    buttons=list([
                        dict(count=1, label="1m",
                             step="month", stepmode="backward"),
                        dict(count=6, label="6m",
                             step="month", stepmode="backward"),
                        dict(count=1, label="YTD",
                             step="year", stepmode="todate"),
                        dict(count=1, label="1y",
                             step="year", stepmode="backward"),
                        dict(step="all")
                    ])
                ),
                rangeslider=dict(
                    visible=True,
                    thickness = 0.05
                ),
                autorange = False,
                range=[df.index[0], df.index[-1]],
                fixedrange= False,
                type="date"
            )
        )

        fig.update_layout(
            yaxis=dict(
                autorange = False,
                range=[df.min().min() * 0.95, df.max().max() * 1.05],
                fixedrange= False
            )
        )

        fig.update_layout(
            autosize=False,
            width=900,
            height=550,
            margin=dict(l=50, r=50, b=25, t=100, pad=1),
            paper_bgcolor="LightSteelBlue",
        )

        fig.show()

        return fig



import numpy as np
import scipy.linalg as la
import scipy.sparse as sps
import warnings
import time

from ._RiskAnalyzer import _RiskAnalyzer
from ._solvers import _socp_solver, _tol_cholesky

class SDAnalyzer(_RiskAnalyzer):
    """
    SD - Standard Deviation based optimal portfolio strategies.
    
    Methods:
        * getWeights
        * getRisk
        * getPositions
        * getRiskComp
        * getDiversification
        * viewForntiers
        * set_rrate
        * set_mktdata
        * set_rtype
        * set_random_seed
    """
    def __init__(self, mktdata=None, colname='adjusted', freq='Q', 
                 hlength=3.25, calendar=None, rtype='Sharpe', method = 'ecos',
                 name='SD'):
        """
        Constructor

        Parameters
        ----------
        `mktdata` : `pandas.DataFrame`, optional
            Historic daily market data for portfolio components in the format
            returned by `azapy.mktData` function. The default is `None`.
        `colname` : `str`, optional;
            Name of the price column from mktdata used in the weights 
            calibration. The default is `'adjusted'`.
        `freq` : `str`, optional;
            Rate of return horizon. It could be 
            `'Q'` for quarter or `'M'` for month. The default is `'Q'`.
        `hlength` : `float`, optional;
            History length in number of years used for calibration. A 
            fractional number will be rounded to an integer number of months.
            The default is `3.25` years.
        `calendar` : `numpy.busdaycalendar`, optional;
            Business days calendar. If is it `None` then the calendar will 
            be set to NYSE business calendar. 
            The default is `None`.
        `rtype` : `str`, optional;
            Optimization type. Possible values: \n
                `'Risk'` : optimal-risk portfolio for targeted expected rate of 
                return.\n
                `'Sharpe'` : Sharpe-optimal portfolio - maximization solution.\n
                `'Sharpe2'` : Sharpe-optimal portfolio - minimization solution.\n
                `'MinRisk'` : minimum risk portfolio.\n
                `'RiskAverse'` : optimal-risk portfolio for a fixed 
                risk-aversion factor.\n
                `'InvNrisk'` : optimal-risk portfolio with the same risk value 
                as a benchmark portfolio (e.g., same as equal weighted 
                portfolio).\n
                `'Diverse'` : optimal-diversified portfolio for targeted
                expected rate of return (maximum of inverse 1-D).\n
                `'Diverse2'` : optimal-diversified portfolio for targeted
                expected rate of return (minmum of 1-D).\n
                `'MaxDiverse'` : maximum diversified portfolio.\n
                `'InvNdiverse'` : optimal-diversified portfolio with the same
                diversification factor as a benchmark portfolio 
                (e.g., same as equal weighted portfolio).\n
                `'InvNdrr'` : optima- diversified portfolio with the same 
                expected rate of return as a benchmark portfolio
                (e.g., same as equal weighted portfolio).\n
            The defauls is `'Sharpe'`.
        `method` : `str`, optional;
            Quadratic programming numerical method. Could be `'ecos'` or
            `'cvxopt'`. The default is `'ecos'`.
        `name` : `str`, optional;
            Object name. The default is `'SD'`.
            
        Returns
        -------
        The object.
        """
        super().__init__(mktdata, colname, freq, hlength, calendar, 
                         rtype, name)
        
        self._set_method(method)
        
        
    def _set_method(self, method):
        self._set_qp_method(method)
        
        
    def _risk_calc(self, prate, alpha):
        var = np.var(prate)
        
        # status, variance, volatility
        return 0, var, np.sqrt(var)
    
    
    def _risk_min(self, d=1):
        # Computes the minimization of volatility
        # Order of variables
        # w <- [0:nn]
        # t <- nn
        # in total dim = nn + 1
        
        # build P
        P = self.rrate.cov().to_numpy()
        nn = P.shape[0]
        
        # build c
        c_data = [0.] * nn + [1.]
        
        # build G
        # linear + first line cone
        icol = list(range(nn)) + list(range(nn)) + [nn]
        irow = [0] * nn + list(range(1, nn + 1)) + [nn + 1]
        data = list(-self.muk * d) + [-1.] * (nn + 1)
        dd = sps.coo_matrix((data, (irow, icol)), shape=(nn + 2, nn + 1))
        
        if any(np.diag(P) < _tol_cholesky):
            pp = np.concatenate((-la.sqrtm(P), np.zeros((nn, 1))), axis=1)
        else:
            pp = np.concatenate((-la.cholesky(P, overwrite_a=True), 
                                 np.zeros((nn, 1))), axis=1)
            
        G = sps.vstack( [dd, pp])
        
        # biuld dims
        dims = {'l': nn + 1, 'q': [nn + 1]}
        
        # build h
        h_data = [-self.mu * d] + [0.] * (nn * 2 + 1)
        
        # build A
        A = sps.coo_matrix([1.] * nn + [0.])
        
        # build
        b_data = [1.]
        
        # calc
        toc = time.perf_counter()
        res = _socp_solver(self.method, c_data, G, h_data, dims, A, b_data)
        self.time_level2 = time.perf_counter() - toc
        
        self.status = res['status']
        if self.status != 0:
            warnings.warn(f"Warning {self.name} on {self.rrate.index[-1]} :: "
                          f"status {res['status']} :: {res['infostring']}")
            return np.array([np.nan] * nn)

        # optimal weights
        self.ww = np.array(res['x'][:-1])
        self.ww.shape = nn
        # volatility
        self.risk = res['pcost']
        # volatility
        self.primary_risk_comp = np.array([self.risk])
        # variance
        self.secondary_risk_comp = np.array([self.risk**2])
        # rate of return
        self.RR = np.dot(self.ww, self.muk)
        
        return self.ww
    
    
    def _sharpe_inv_min(self):
        # Computes the minimization of the inverse of Sharpe
        # Order of variables
        # w <- [0:nn]
        # t <- nn
        # u <- nn + 1
        # in total dim = nn + 2
        P = self.rrate.cov().to_numpy()
        nn = P.shape[0]
        
        # build c
        c_data = [0.] * (nn + 1) + [1.]
       
        # biuld G
        dd = sps.block_diag((sps.diags([-1.] * nn, format='coo'), [0.,-1.]))
        
        if any(np.diag(P) < _tol_cholesky):
            pp = np.concatenate((-la.sqrtm(P), np.zeros((nn,2))), axis=1)
        else:
            pp = np.concatenate((-la.cholesky(P, overwrite_a=True), 
                                 np.zeros((nn,2))), axis=1)
                
        G = sps.vstack([dd, pp])
        
        # biuld dims
        dims = {'l': nn, 'q': [nn + 1]}
        
        # build h
        h_data = [0.] * (2 * nn + 1)
        
        # build A
        A = sps.coo_matrix(
            [list(self.muk) + [-self.mu] + [0.], [1.] * nn + [-1.] + [0.]])
        
        # build b
        b_data = [1., 0.]
        
        # calc
        toc = time.perf_counter()
        res = _socp_solver(self.method, c_data, G, h_data, dims, A, b_data)
        self.time_level2 = time.perf_counter() - toc
        
        self.status = res['status']
        if self.status != 0:
            warnings.warn(f"Warning {self.name} on {self.rrate.index[-1]} :: "
                          f"status {res['status']} :: {res['infostring']}")
            return np.array([np.nan] * nn)

        t = res['x'][-2]
        # sharpe
        self.sharpe = 1. / res['pcost']
        # optimal weights
        self.ww = np.array(res['x'][:nn]) / t
        self.ww.shape = nn
        # rate of return
        self.RR = self.mu + 1. / t
        # volatility
        self.risk = res['pcost'] / t
        # volatility
        self.primary_risk_comp = np.array([self.risk])
        # variance
        self.secondary_risk_comp = np.array([self.risk**2])
        
        return self.ww
    
    
    def _sharpe_max(self):
        # Computes the maximization of Sharpe
        # Order of variables
        # w <- [0:nn]
        # t <- nn
        # in total dim = nn + 1
        P = self.rrate.cov().to_numpy()
        nn = P.shape[0]
        
        # build c
        c_data = list(-self.muk) + [self.mu]

        # build G
        dd = sps.diags([-1.] * nn + [0.], format='coo')
        
        if any(np.diag(P) < _tol_cholesky):
            pp = np.concatenate((-la.sqrtm(P), np.zeros((nn, 1))), axis=1)
        else:
            pp = np.concatenate((-la.cholesky(P, overwrite_a=True), 
                                 np.zeros((nn, 1))), axis=1)
            
        G = sps.vstack([dd, pp])

        # build h
        h_data = [0.] * nn + [1.] + [0.] * nn 
        
        # def dims
        dims = {'l': nn, 'q': [nn + 1]}
        
        # build A
        A = sps.coo_matrix([1.] * nn + [-1.])
        
        # build b
        b_data = np.array([0.])
        
        # calc
        toc = time.perf_counter()
        res = _socp_solver(self.method, c_data, G, h_data, dims, A, b_data)
        self.time_level2 = time.perf_counter() - toc
 
        self.status = res['status']
        if self.status != 0:
            warnings.warn(f"Warning {self.name} on {self.rrate.index[-1]} :: "
                          f"status {res['status']} :: {res['infostring']}")
            return np.array([np.nan] * nn)
 
        t = res['x'][-1]
        # optimal weights
        self.ww = np.array(res['x'][:nn]) / t
        self.ww.shape = nn
        # sharpe
        self.sharpe = -res['pcost']
        # volatility
        self.risk = 1. / t
        # volatility
        self.primary_risk_comp = np.array([self.risk])
        # variance
        self.secondary_risk_comp = np.array([self.risk**2])
        # rate of return
        self.RR = self.sharpe / t + self.mu
        
        return self.ww   
    
    
    def _rr_max(self):
        # Computes the maximization of returns (for fixed volatility)
        # Order of variables
        # w <- [0:nn]
        # in total dim = nn 
        P = self.rrate.cov().to_numpy()
        nn = P.shape[0]
        
        # build c
        c_data = list(-self.muk)
        
        # build G
        # ww >= 0 + 1st line in cone
        dd = sps.diags([-1.] * nn, shape=(nn + 1, nn), format='coo')
        
        if any(np.diag(P) < _tol_cholesky):
            G = sps.vstack([dd, -la.sqrtm(P)])
        else:
            G = sps.vstack([dd, -la.cholesky(P, overwrite_a=True)])
        
        # build h
        h_data = [0.] * nn + [self.risk] + [0.] * nn
        
        # def dims
        dims = {'l': nn, 'q': [nn + 1]}
        
        # build A_eq
        A = sps.coo_matrix([1.] * nn) 
        
        # build b_eq
        b_data = [1.]
        
        # calc
        toc = time.perf_counter()
        res = _socp_solver(self.method, c_data, G, h_data, dims, A, b_data)
        self.time_level2 = time.perf_counter() - toc
 
        self.status = res['status']
        if self.status != 0:
            warnings.warn(f"Warning {self.name} on {self.rrate.index[-1]} :: "
                          f"status {res['status']} :: {res['infostring']}")
            return np.array([np.nan] * nn)

        # optimal weights
        self.ww = np.array(res['x'][:])
        self.ww.shape = nn
        # rate of return
        self.RR = -res['pcost']
        # volatility
        self.primary_risk_comp = np.array([self.risk])
        # variance
        self.secondary_risk_comp = np.array([self.risk**2])
        
        return self.ww   
    
    
    def _risk_averse(self):
        # Order of variables
        # w <- [0:nn]
        # t <- nn
        # in total dim = nn + 1
        
        # build P
        P = self.rrate.cov().to_numpy() 
        nn = P.shape[0]
        
        # build c
        c_data = list(-self.muk) + [self.Lambda]
        
        # build G
        dd = sps.diags([-1.] * (nn + 1))
        
        if any(np.diag(P) < _tol_cholesky):
            pp = np.concatenate((-la.sqrtm(P), np.zeros((nn,1))), axis=1)
        else:
            pp = np.concatenate((-la.cholesky(P, overwrite_a=True), 
                                 np.zeros((nn,1))), axis=1)
            
        G = sps.vstack([dd, pp])
        
        # build dims
        dims = {'l': nn, 'q': [nn + 1]}
        
        # build h
        h_data = [0.] * (2 * nn + 1)
        
        # build A
        A = sps.coo_matrix([1.] * nn + [0.])
        
        # build
        b_data = [1.]
        
        # calc
        toc = time.perf_counter()
        res = _socp_solver(self.method, c_data, G, h_data, dims, A, b_data)
        self.time_level2 = time.perf_counter() - toc
 
        self.status = res['status']
        if self.status != 0:
            warnings.warn(f"Warning {self.name} on {self.rrate.index[-1]} :: "
                          f"status {res['status']} :: {res['infostring']}")
            return np.array([np.nan] * nn)
        
        # optimal weights
        self.ww = np.array(res['x'][:nn])
        self.ww.shape = nn
        # rate of return
        self.RR = np.dot(self.ww, self.muk)
        # volatility
        self.risk = (res['pcost'] + self.RR) / self.Lambda
        # volatility
        self.primary_risk_comp = np.array([self.risk])
        # variance
        self.secondary_risk_comp = np.array([self.risk**2])
        
        return self.ww
    
    
    def _risk_diversification(self, d=1):
        # Computes the minimization of the inverse of Sharpe
        # Order of variables
        # w <- [0:nn]
        # t <- nn
        # u <- nn + 1
        # in total dim = nn + 2
        P = self.rrate.cov().to_numpy()
        nn = P.shape[0]
        
        # build c
        c_data = [0.] * (nn + 1) + [1.]
       
        # biuld G
        dd = sps.block_diag((sps.diags([-1.] * nn, format='coo'), [0., -1.]))
 
        if any(np.diag(P) < _tol_cholesky):
            pp = np.concatenate((-la.sqrtm(P), np.zeros((nn,2))), axis=1)
        else:
            pp = np.concatenate((-la.cholesky(P, overwrite_a=True), 
                                 np.zeros((nn,2))), axis=1)
          
        xx = sps.coo_matrix(list(-self.muk * d) + [self.mu * d, 0.])
        
        G = sps.vstack([xx, dd, pp])
        
        # biuld dims
        dims = {'l': nn + 1, 'q': [nn + 1]}
        
        # build h
        h_data = [0.] * (2 * nn + 2)
        
        # build A
        A = sps.coo_matrix(
            [list(self.risk_comp) + [0.] + [0.], [1.] * nn + [-1.] + [0.]])
        
        # build b
        b_data = [1., 0.]
        
        # calc
        toc = time.perf_counter()
        res = _socp_solver(self.method, c_data, G, h_data, dims, A, b_data)
        self.time_level2 = time.perf_counter() - toc
        
        self.status = res['status']
        if self.status != 0:
            warnings.warn(f"Warning {self.name} on {self.rrate.index[-1]} :: "
                          f"status {res['status']} :: {res['infostring']}")
            return np.array([np.nan] * nn)

        t = res['x'][-2]
        # SD-Divers
        self.diverse= 1. - res['pcost']
        # optimal weights
        self.ww = np.array(res['x'][:nn]) / t
        self.ww.shape = nn
        # rate of return
        self.RR = np.dot(self.ww, self.muk)
        # volatility
        self.risk = res['pcost'] / t
        # volatility
        self.primary_risk_comp = np.array([self.risk])
        # variance
        self.secondary_risk_comp = np.array([self.risk**2])
        
        return self.ww
    
    
    def _risk_inv_diversification(self, d=1):
        # Computes the minimization of the inverse of Sharpe
        # Order of variables
        # w <- [0:nn]
        #   t <- nn
        # in total dim = nn + 1
        P = self.rrate.cov().to_numpy()
        nn = P.shape[0]
        
        # build c
        c_data = list(-self.risk_comp) + [0.] 
        
        # build G
        # linear + first line cone
        icol = list(range(nn + 1)) + list(range(nn + 1))
        irow = [0] * (nn + 1) + list(range(1, nn + 2)) 
        data = list(-self.muk * d) + [self.mu * d] + [-1.] * (nn + 1)
        dd = sps.coo_matrix((data, (irow, icol)), shape=(nn + 3, nn + 1))
        
        if any(np.diag(P) < _tol_cholesky):
            pp = np.concatenate((-la.sqrtm(P), np.zeros((nn, 1))), axis=1)
        else:
            pp = np.concatenate((-la.cholesky(P, overwrite_a=True), 
                                 np.zeros((nn, 1))), axis=1)
            
        G = sps.vstack([dd, pp])

        # build h
        h_data = [0.] * (nn + 2) + [1.] + [0.] * nn 
        
        # def dims
        dims = {'l': nn + 2, 'q': [nn + 1]}
        
        # build A
        A = sps.coo_matrix([1.] * nn + [-1.])
        
        # build b
        b_data = np.array([0.])
        
        # calc
        toc = time.perf_counter()
        self.time_level2 = time.perf_counter() - toc
        res = _socp_solver(self.method, c_data, G, h_data, dims, A, b_data)
        
        self.status = res['status']
        if self.status != 0:
            warnings.warn(f"Warning {self.name} on {self.rrate.index[-1]} :: "
                          f"status {res['status']} :: {res['infostring']}")
            return np.array([np.nan] * nn)

        t = res['x'][-1]
        # SD-Divers
        self.diverse= 1. + 1. / res['pcost']
        # optimal weights
        self.ww = np.array(res['x'][:nn]) / t
        self.ww.shape = nn
        # rate of return
        self.RR = np.dot(self.ww, self.muk)
        # volatility
        self.risk = 1. / t
        # volatility
        self.primary_risk_comp = np.array([self.risk])
        # variance
        self.secondary_risk_comp = np.array([self.risk**2])
        
        return self.ww
        
    
    def _rr_max_diversification(self):
        # Computes the maximization of returns (for fixed volatility)
        # Order of variables
        # w <- [0:nn]
        # in total dim = nn 
        P = self.rrate.cov().to_numpy()
        nn = P.shape[0]
        
        # build c
        c_data = list(-self.muk)
        
        # build G
        # ww >= 0 
        dd = sps.diags([-1.] * nn, shape=(nn, nn), format='coo')
        
        # cone
        xx = sps.coo_matrix(list((self.diverse- 1) * self.risk_comp), shape=(1, nn))
        
        if any(np.diag(P) < _tol_cholesky):
            G = sps.vstack([dd, xx,  -la.sqrtm(P)])
        else:
            G = sps.vstack([dd, xx, -la.cholesky(P, overwrite_a=True)])
        
        # build h
        h_data = [0.] * (2 * nn  + 1)
        
        # def dims
        dims = {'l': nn, 'q': [nn + 1]}
        
        # build A_eq
        A = sps.coo_matrix([1.] * nn) 
        
        # build b_eq
        b_data = [1.]
        
        # calc
        toc = time.perf_counter()
        res = _socp_solver(self.method, c_data, G, h_data, dims, A, b_data)
        self.time_level2 = time.perf_counter() - toc
 
        self.status = res['status']
        if self.status != 0:
            warnings.warn(f"Warning {self.name} on {self.rrate.index[-1]} :: "
                          f"status {res['status']} :: {res['infostring']}")
            return np.array([np.nan] * nn)

        # optimal weights
        self.ww = np.array(res['x'][:])
        self.ww.shape = nn
        # rate of return
        self.RR = -res['pcost']
        # risk
        self.risk = np.sqrt(np.dot(np.dot(self.ww, P), self.ww))
        # volatility
        self.primary_risk_comp = np.array([self.risk])
        # variance
        self.secondary_risk_comp = np.array([self.risk**2])
        # diversification
        self.diverse = 1 - self.risk / np.dot(self.ww, self.risk_comp)
        
        return self.ww   


from .Port_MV import Port_MV
from .SDAnalyzer import SDAnalyzer

class Port_SD(Port_MV):
    """
    Backtesting the SD optimal portfolio periodically rebalanced.
    
    Methods:
        * set_model
        * get_port
        * get_nshares
        * get_weights
        * get_account
        * get_mktdata
        * port_view
        * port_view_all
        * port_drawdown
        * port_perf
        * port_annual_returns
        * port_monthly_returns
        * port_period_returns
    """
    def _wwgen(self):
        return SDAnalyzer(rtype=self.rtype, method=self.method,
                          name=self.pname)
    


import pandas as pd
import numpy as np

from .Port_Rebalanced import Port_Rebalanced
from azapy.MkT.MkTcalendar import NYSEgen
from azapy.util.schedule import schedule_roll

class Port_ConstW(Port_Rebalanced):
    """
    Backtesting portfolio with constant weights, periodically rebalanced.

    Methods:
        * set_model
        * get_port
        * get_nshares
        * get_weights
        * get_account
        * get_mktdata
        * port_view
        * port_view_all
        * port_drawdown
        * port_perf
        * port_annual_returns
        * port_monthly_returns
        * port_period_returns
    """
    def __init__(self, mktdata, symb=None, sdate=None, edate=None,
                 col_price='close', col_divd='divd', col_ref='adjusted',
                 pname='Port', pcolname=None, capital=100000,
                 schedule=None,
                 freq='Q', noffset=-3, fixoffset=-1, calendar=None):
        """
        Constructor

        Parameters
        ----------
        `mktdata` : `pandas.DataFrame`;
            MkT data in the format "symbol", "date", "open", "high", "low",
            "close", "volume", "adjusted", "divd", "split" (e.g. as returned
            by `azapy.readMkT` function).
        `symb` : `list`, optional;
            List of symbols for the basket components. All symbols MkT data
            should be included in mktdata. If set to `None` the `symb` will be
            set to include all the symbols from `mktdata`. The default
            is `None`.
        `sdate` : date like, optional;
            Start date for historical data. If set to `None` the `sdate` will
            be set to the earliest date in mktdata. The default is `None`.
        `edate` : date like, optional;
            End date for historical dates and so the simulation. Must be
            greater than  `sdate`. If it is `None` then `edat`e will be set
            to the latest date in mktdata. The default is `None`.
        `col_price` : `str`, optional;
            Column name in the mktdata DataFrame that will be considered
            for portfolio aggregation. The default is `'close'`.
        `col_divd` :  `str`, optional;
            Column name in the mktdata DataFrame that holds the dividend
            information. The default is `'dvid'`.
        `col_ref` : `str`, optional;
            Column name in the mktdata DataFrame that will be used as a price
            reference for portfolio components. The default is `'adjusted'`.
        `pname` : `str`, optional;
            The name of the portfolio. The default is `'Port'`.
        `pcolname` : `str`, optional;
            Name of the portfolio price column. If it set to `None` that
            `pcolname=pname`. The default is `None`.
        `capital` : `float`, optional;
            Initial portfolio Capital in dollars. The default is `100000`.
        `schedule` : `pandas.DataFrame`, optional;
            Rebalancing schedule, with columns for `'Droll'` rolling date and
            `'Dfix'` fixing date. If it is `None` than the schedule will be set
            using the `freq`, `noffset`, `fixoffset` and `calendar`
            information. The default is `None`.
        `freq` : `str`, optional;
            rebalancing frequency. It can be `'Q'` for quarterly or `'M'` for
            monthly rebalancing, respectively. It is relevant only is schedule
            is `None`. The default is `'Q'`.
        `noffset` : `int`, optional;
            Number of business days offset for rebalancing date `'Droll'`
            relative to the end of the period (quart or month). A positive
            value add business days beyond the calendar end of the period while
            a negative value subtracts business days. It is relevant only is
            schedule is `None`. The default is `-3`.
        `fixoffset` : `int`, optional;
            Number of business day offset of fixing date `'Dfix'` relative to
            the rebalancing date `'Droll'`. It can be 0 or negative. It is
            relevant only is schedule is `None`. The default is `-1`.
        `calendar` : `numpy.busdaycalendar`, optional;
            Business calendar. If it is `None` then it will be set to NYSE
            business calendar. The default
            vale is `None`.

        Returns
        -------
        The object.
        """
        super().__init__(mktdata=mktdata, symb=symb,
                         sdate=sdate, edate=edate,
                         col_price=col_price, col_divd=col_divd,
                         col_ref=col_ref,
                         pname=pname,
                         pcolname=pcolname, capital=capital)
        self.schedule = schedule
        self.freq = freq
        self.noffset = noffset
        self.fixoffset = fixoffset
        self.calendar =  calendar
        if self.calendar is None: self._default_calendar()


    def set_model(self, ww=None):
        """
        Set model parameters and evaluate the portfolio time-series.

        Parameters
        ----------
        `ww` : `list` (also `numpy.array` or `pandas.Series`), optional

            List of weights. If it is `pandas.Series` the index should match
            the basket `symb`. Otherwise, the weights are considered in the 
            `symb` order. 
            If it is set to `None` than `ww` will be set to equal weights.
            The default is `None`.

        Returns
        -------
        `pands.DataFrame`
            The portfolio time-series in the format "date", "pcolname".
        """
        self.hlength = 0.

        self._set_schedule()
        self._set_weights(ww)
        self._port_calc()
        return self.port

    def _default_calendar(self):
        self.calendar = NYSEgen()

    def _set_schedule(self):
        if self.schedule is None:
            self.schedule = schedule_roll(self.sdate, self.edate, self.freq,
                                          self.noffset, self.fixoffset,
                                          self.calendar, self.hlength)

    def _set_weights(self, ww):
        if ww is None:
            _ww = pd.Series(1., index=self.symb)
        elif isinstance(ww, pd.core.series.Series):
            _ww = ww
        else:
            _ww = pd.Series(ww, index=self.symb)

        if _ww.size != self.symb.size:
            raise ValueError(f"ww wrong size, it must be {self.symb.size}")

        if np.any(_ww < 0.):
            raise ValueError("All ww elements must be >= 0")

        wws = _ww.sum()
        if wws <= 0.:
            raise ValueError("At least one ww element must be > 0")

        self.ww = self.schedule
        for sy in self.symb:
            self.ww[sy] = _ww[sy] / wws



import numpy as np
import scipy.sparse as sps
import warnings
import time

from ._RiskAnalyzer import _RiskAnalyzer
from ._solvers import _lp_solver

class GINIAnalyzer(_RiskAnalyzer):
    """
    GINI based optimal portfolio strategies.
    
    Methods:
        * getWeights
        * getRisk
        * getPositions
        * getRiskComp
        * getDiversification
        * viewForntiers
        * set_rrate
        * set_mktdata
        * set_rtype
        * set_random_seed
    """
    def __init__(self, mktdata=None, colname='adjusted', freq='Q', 
                 hlength=1.25, calendar=None, 
                 rtype='Sharpe', method='ecos', name='GINI'):
        """
        Constructor

        Parameters
        ----------
        `mktdata` : `pandas.DataFrame`, optional;
            Historic daily market data for portfolio components in the format
            returned by `azapy.mktData` function. The default is `None`.
        `colname` : `str`, optional;
            Name of the price column from mktdata used in the weights 
            calibration. The default is `'adjusted'`.
        `freq` : `str`, optional;
            Rate of return horizon. It could be 
            `'Q'` for quarter or `'M'` for month. The default is `'Q'`.
        `hlength` : `float`, optional;
            History length in number of years used for calibration. A 
            fractional number will be rounded to an integer number of months.
            The default is `1.25` years.
        `calendar` : `numpy.busdaycalendar`, optional;
            Business days calendar. If is it `None` then the calendar will 
            be set to NYSE business calendar. 
            The default is `None`.
        `rtype` : `str`, optional;
            Optimization type. Possible values: \n
                `'Risk'` : optimal-risk portfolio for targeted expected rate of 
                return.\n
                `'Sharpe'` : Sharpe-optimal portfolio - maximization solution.\n
                `'Sharpe2'` : Sharpe-optimal portfolio - minimization solution.\n
                `'MinRisk'` : minimum risk portfolio.\n
                `'RiskAverse'` : optimal-risk portfolio for a fixed 
                risk-aversion factor.\n
                `'InvNrisk'` : optimal-risk portfolio with the same risk value 
                as a benchmark portfolio (e.g., same as equal weighted 
                portfolio).\n
                `'Diverse'` : optimal-diversified portfolio for targeted
                expected rate of return (maximum of inverse 1-D).\n
                `'Diverse2'` : optimal-diversified portfolio for targeted
                expected rate of return (minmum of 1-D).\n
                `'MaxDiverse'` : maximum diversified portfolio.\n
                `'InvNdiverse'` : optimal-diversified portfolio with the same
                diversification factor as a benchmark portfolio 
                (e.g., same as equal weighted portfolio).\n
                `'InvNdrr'` : optima- diversified portfolio with the same 
                expected rate of return as a benchmark portfolio
                (e.g., same as equal weighted portfolio).\n
            The defauls is `'Sharpe'`.
        `method` : `str`, optional;
            Linear programming numerical method. 
            Could be: `'ecos'`, `'highs-ds'`, `'highs-ipm'`, `'highs'`, 
            `'interior-point'`, `'glpk'` and `'cvxopt'`.
            The default is `'ecos'`.
        `name` : `str`, optional;
            Object name. The default is `'GINI'`.
            
        Returns
        -------
        The object.
        """
        self.drate = None
        self.nn2 = None
        super().__init__(mktdata, colname, freq, hlength, calendar, rtype,
                         name)
        
        self._set_method(method)
        
        
    def _set_method(self, method):
        self._set_lp_method(method)
        
    
    def set_rrate(self, rrate):
        """
        Sets the MkT Data.

        Parameters
        ----------
        rrate : `pandas.DataFrame`
            Market data. It will overwrite the value set by the constructor.
        Returns
        -------
        `None`
        """
        self.nn, self.mm = rrate.shape
        self.muk = rrate.mean()
        self.rrate = rrate.copy()
        
        self.nn2 = int(self.nn * (self.nn - 1) / 2)
        yy = []
        for m in range(self.mm):
            x = self.rrate.iloc[:,m]
            yy.append([x[i] - x[j] \
                       for i in range(self.nn - 1) \
                       for j in range(i + 1, self.nn)])
        self.drate = np.concatenate(yy)
        
        
    def _risk_calc(self, prate, alpha):
        nn = len(prate)
 
        y = np.sort(prate, kind='heapsort')
        gini = np.sum([y[i] * ( 2 * i - nn + 1) for i in range(nn)]) / nn / nn
        # status, gini, gini
        return 0, gini, gini
    
    
    def _risk_min(self, d=1):
        # Order of variables (mm - no of symb, nn - no of observations)
        # w <- [0 : mm]
        # s <- [mm : mm + nn2]
        # in total dim = mm + nn2 
        # where nn2 = nn * (nn - 1) / 2
        mm = self.mm
        nn2 = self.nn2
        norm = 1 / self.nn ** 2
        
        # build c
        c_data = [0.] * mm + [norm] * nn2
        
        # bild G
        G_icol = [m for m in range(mm) for _ in range(nn2)] * 2
        G_irow = list(range(nn2)) * mm + list(range(nn2, 2 * nn2)) * mm
        G_data = list(self.drate) + list(-self.drate)
        
        G_icol += list(range(mm, mm + nn2)) * 2
        G_irow += list(range(nn2 * 2))
        G_data += [-1.] * (nn2 * 2)
        
        G_icol += list(range(mm))
        G_irow += [nn2 * 2] * mm
        G_data += list(-self.muk * d)
        
        G_icol += list(range(mm + nn2))
        G_irow += list(range(nn2 * 2 + 1, nn2 * 3 + 1 + mm))
        G_data += [-1.] * (mm + nn2)
        
        G_shape = (nn2 * 3 + 1 + mm, mm + nn2)
        G = sps.coo_matrix((G_data, (G_irow, G_icol)), G_shape)
   
        # build h
        h_data = [0.] * (nn2 * 2) + [-self.mu * d] + [0.] * (mm + nn2)
        
        # build A
        A_icol = list(range(mm))
        A_irow = [0] * mm
        A_data = [1.] * mm
        A_shape = (1, mm + nn2)
        A = sps.coo_matrix((A_data, (A_irow, A_icol)), A_shape)
        
        # build b
        b_data = [1.]
        
        # calc
        toc = time.perf_counter()
        res = _lp_solver(self.method, c_data, G, h_data, A, b_data)
        self.time_level2 = time.perf_counter() - toc
 
        self.status = res['status']
        if self.status != 0:
            warnings.warn(f"Warning {self.name} on {self.rrate.index[-1]} :: "
                          f"status {res['status']} :: {res['infostring']}")
            return np.array([np.nan] * mm)
            
        # GINI
        self.risk = res['pcost']
        # optimal weights
        self.ww = np.array(res['x'][:mm])
        self.ww.shape = mm
        # rate of return
        self.RR = np.dot(self.ww, self.muk)
        # defalut to risk
        self.primary_risk_comp = np.array([self.risk])
        self.secondary_risk_comp = np.array([self.risk])
        
        return self.ww
    
    
    def _sharpe_max(self):
        # Order of variables (mm - no of symb, nn - no of observations)
        # w <- [0 : mm]
        # s <- [mm : mm + nn2]
        # t <- mm + nn2
        # in total dim = mm + nn2 + 1
        # where nn2 = nn * (nn - 1) / 2
        mm = self.mm
        nn2 = self.nn2
        norm = 1 / self.nn ** 2
        
        # build c
        c_data = list(-self.muk) + [0.] * nn2 + [self.mu]
        
        # bild G
        G_icol = [m for m in range(mm) for _ in range(nn2)] * 2
        G_irow = list(range(nn2)) * mm + list(range(nn2, 2 * nn2)) * mm
        G_data = list(self.drate) + list(-self.drate)
        
        G_icol += list(range(mm, mm + nn2)) * 2
        G_irow += list(range(nn2 * 2))
        G_data += [-1.] * (nn2 * 2)
        
        G_icol += list(range(mm))
        G_irow += list(range(nn2 * 2, nn2 * 2 + mm))
        G_data += [-1.] * mm
        
        G_shape = (nn2 * 2 + mm , mm + nn2 + 1)
        G = sps.coo_matrix((G_data, (G_irow, G_icol)), G_shape)
        
        # build h
        h_data = [0.] * (nn2 * 2 + mm) 
        
        # build A
        A_icol = list(range(mm, mm + nn2)) + list(range(mm)) + [mm + nn2] 
        A_irow = [0] * nn2 + [1] * (mm + 1)
        A_data = [norm] * nn2 + [1.] * mm + [-1.]
        
        A_shape = (2, mm + nn2 + 1)
        A = sps.coo_matrix((A_data, (A_irow, A_icol)), A_shape)
        
        # build b
        b_data = [1., 0.]
        
        # calc
        toc = time.perf_counter()
        res = _lp_solver(self.method, c_data, G, h_data, A, b_data)
        self.time_level2 = time.perf_counter() - toc
 
        self.status = res['status']
        if self.status != 0:
            warnings.warn(f"Warning {self.name} on {self.rrate.index[-1]} :: "
                          f"status {res['status']} :: {res['infostring']}")
            return np.array([np.nan] * mm)
            
        t = res['x'][-1]
        # Gini-Sharpe
        self.sharpe = -res['pcost']
        # GINI
        self.risk = 1. / t
        # optimal weights
        self.ww = np.array(res['x'][:mm] / t)
        self.ww.shape = mm
        # rate of return
        self.RR = -res['pcost'] / t + self.mu
        # default to risk
        self.primary_risk_comp = np.array([self.risk])
        self.secondary_risk_comp = np.array([self.risk])
        
        return self.ww
    
    
    def _sharpe_inv_min(self):
        # Order of variables (mm - no of symb, nn - no of observations)
        # w <- [0 : mm]
        # s <- [mm : mm + nn2]
        # t <- mm + nn2
        # in total dim = mm + nn2 + 1
        # where nn2 = nn * (nn - 1) / 2
        mm = self.mm
        nn2 = self.nn2
        norm = 1 / self.nn ** 2
        
        # build c
        c_data = [0.] * mm + [norm] * nn2 + [0.]
        
        # build G
        G_icol = [m for m in range(mm) for _ in range(nn2)] * 2
        G_irow = list(range(nn2)) * mm + list(range(nn2, 2 * nn2)) * mm
        G_data = list(self.drate) + list(-self.drate)
        
        G_icol += list(range(mm, mm + nn2)) * 2
        G_irow += list(range(nn2 * 2))
        G_data += [-1.] * (nn2 * 2)
        
        G_icol += list(range(mm))
        G_irow += list(range(nn2 * 2, nn2 * 2 + mm))
        G_data += [-1.] * mm
        
        G_shape = (nn2 * 2 + mm, mm + nn2 + 1)
        G = sps.coo_matrix((G_data, (G_irow, G_icol)), G_shape)
        
        # build h
        h_data = [0.] * (nn2 * 2 + mm) 
        
        # build A
        A_icol = (list(range(mm)) + [mm + nn2]) * 2
        A_irow = [0] * (mm + 1) + [1] * (mm + 1)
        A_data = list(self.muk) + [-self.mu] + [1.] * mm + [-1.]
        
        A_shape = (2, mm + nn2 + 1)
        A = sps.coo_matrix((A_data, (A_irow, A_icol)), A_shape)
        
        # build b
        b_data = [1., 0.]
        
        # calc
        toc = time.perf_counter()
        res = _lp_solver(self.method, c_data, G, h_data, A, b_data)
        self.time_level2 = time.perf_counter() - toc
 
        self.status = res['status']
        if self.status != 0:
            warnings.warn(f"Warning {self.name} on {self.rrate.index[-1]} :: "
                          f"status {res['status']} :: {res['infostring']}")
            return np.array([np.nan] * mm)
            
        t = res['x'][-1]
        # Gini-Sharpe
        self.sharpe = 1. / res['pcost']
        # GINI
        self.risk = res['pcost'] / t
        # optimal weights
        self.ww = np.array(res['x'][:mm] / t)
        self.ww.shape = mm
        # rate of return
        self.RR = 1. / t + self.mu
        # default to risk
        self.primary_risk_comp = np.array([self.risk])
        self.secondary_risk_comp = np.array([self.risk])
        
        return self.ww
    
    
    def _rr_max(self):
        # Order of variables (mm - no of symb, nn - no of observations)
        # w <- [0 : mm]
        # s <- [mm : mm + nn2]
        # in total dim = mm + nn2
        # where nn2 = nn * (nn - 1) / 2
        mm = self.mm
        nn2 = self.nn2
        norm = 1 / self.nn ** 2
        
        # build c
        c_data = list(-self.muk) + [0.] * nn2
        
        # build G
        G_icol = [m for m in range(mm) for _ in range(nn2)] * 2
        G_irow = list(range(nn2)) * mm + list(range(nn2, 2 * nn2)) * mm
        G_data = list(self.drate) + list(-self.drate)
        
        G_icol += list(range(mm, mm + nn2)) * 2
        G_irow += list(range(nn2 * 2))
        G_data += [-1.] * (nn2 * 2)
        
        G_icol += list(range(mm))
        G_irow += list(range(nn2 * 2, nn2 * 2 + mm))
        G_data += [-1.] * mm
        
        G_shape = (nn2 * 2 + mm, mm + nn2)
        G = sps.coo_matrix((G_data, (G_irow, G_icol)), G_shape)
        
        # build h
        h_data = [0.] * (nn2 * 2 + mm) 
        
        # build A
        A_icol = list(range(mm)) + list(range(mm, mm + nn2))
        A_irow = [0] * mm + [1] * nn2
        A_data = [1.] * mm + [norm] * nn2
        
        A_shape = (2, mm + nn2)
        A = sps.coo_matrix((A_data, (A_irow, A_icol)), A_shape)
        
        # build b
        b_data = [1., self.risk]
        
        # calc
        toc = time.perf_counter()
        res = _lp_solver(self.method, c_data, G, h_data, A, b_data)
        self.time_level2 = time.perf_counter() - toc
 
        self.status = res['status']
        if self.status != 0:
            warnings.warn(f"Warning {self.name} on {self.rrate.index[-1]} :: "
                          f"status {res['status']} :: {res['infostring']}")
            return np.array([np.nan] * mm)
            
        # optimal weights
        self.ww = np.array(res['x'][:mm])
        self.ww.shape = mm
        # rate of return
        self.RR = -res['pcost']
        # default to risk
        self.primary_risk_comp = np.array([self.risk])
        self.secondary_risk_comp = np.array([self.risk])
        
        return self.ww


    def _risk_averse(self):
        # Order of variables (mm - no of symb, nn - no of observations)
        # w <- [0 : mm]
        # s <- [mm : mm + nn2]
        # in total dim = mm + nn2 
        # where nn2 = nn * (nn - 1) / 2
        mm = self.mm
        nn2 = self.nn2
        norm = 1 / self.nn ** 2
        
        # build c
        c_data = list(-self.muk) + [self.Lambda * norm] * nn2
        
        # bild G
        G_icol = [m for m in range(mm) for _ in range(nn2)] * 2
        G_irow = list(range(nn2)) * mm + list(range(nn2, 2 * nn2)) * mm
        G_data = list(self.drate) + list(-self.drate)
        
        G_icol += list(range(mm, mm + nn2)) * 2
        G_irow += list(range(nn2 * 2))
        G_data += [-1.] * (nn2 * 2)
        
        G_icol += list(range(mm))
        G_irow += list(range(nn2 * 2, nn2 * 2 + mm))
        G_data += [-1.] * (mm)
        
        G_shape = (nn2 * 2 + mm, mm + nn2)
        G = sps.coo_matrix((G_data, (G_irow, G_icol)), G_shape)
        
        # build h
        h_data = [0.] * (nn2 * 2 + mm)
        
        # build A
        A_icol = list(range(mm))
        A_irow = [0] * mm
        A_data = [1.] * mm
        
        A_shape = (1, mm + nn2)
        A = sps.coo_matrix((A_data, (A_irow, A_icol)), A_shape)
        
        # build b
        b_data = [1.]
        
        # calc
        toc = time.perf_counter()
        res = _lp_solver(self.method, c_data, G, h_data, A, b_data)
        self.time_level2 = time.perf_counter() - toc
 
        self.status = res['status']
        if self.status != 0:
            warnings.warn(f"Warning {self.name} on {self.rrate.index[-1]} :: "
                          f"status {res['status']} :: {res['infostring']}")
            return np.array([np.nan] * mm)
            
        # optimal weights
        self.ww = np.array(res['x'][:mm])
        self.ww.shape = mm
        # rate of return
        self.RR = np.dot(self.ww, self.muk)
        # GINI
        self.risk = (res['pcost'] + self.RR) / self.Lambda
        # default to risk
        self.primary_risk_comp = np.array([self.risk])
        self.secondary_risk_comp = np.array([self.risk])
        
        return self.ww
    
    
    def _risk_diversification(self, d=1):
        # Order of variables (mm - no of symb, nn - no of observations)
        # w <- [0 : mm]
        # s <- [mm : mm + nn2]
        # t <- mm + nn2
        # in total dim = mm + nn2 + 1
        # where nn2 = nn * (nn - 1) / 2
        mm = self.mm
        nn2 = self.nn2
        norm = 1 / self.nn ** 2
        
        # build c
        c_data = [0.] * mm + [norm] * nn2 + [0.]
        
        # build G
        G_icol = [m for m in range(mm) for _ in range(nn2)] * 2
        G_irow = list(range(nn2)) * mm + list(range(nn2, 2 * nn2)) * mm
        G_data = list(self.drate) + list(-self.drate)
        
        G_icol += list(range(mm, mm + nn2)) * 2
        G_irow += list(range(nn2 * 2))
        G_data += [-1.] * (nn2 * 2)
        
        G_icol += list(range(mm)) + [mm + nn2]
        G_irow += [nn2 * 2] * (mm + 1)
        G_data += list(-self.muk * d) + [self.mu * d]
        
        G_icol += list(range(mm))
        G_irow += list(range(nn2 * 2 + 1, nn2 * 2 + mm + 1))
        G_data += [-1.] * mm
        
        G_shape = (nn2 * 2 + mm + 1, mm + nn2 + 1)
        G = sps.coo_matrix((G_data, (G_irow, G_icol)), G_shape)
        
        # build h
        h_data = [0.] * (nn2 * 2 + mm + 1) 
        
        # build A
        A_icol = list(range(mm))
        A_irow = [0] * mm
        A_data = list(self.risk_comp)
        
        A_icol += list(range(mm)) + [mm + nn2]
        A_irow += [1] * (mm + 1)
        A_data += [1.] * mm + [-1.]
        
        A_shape = (2, mm + nn2 + 1)
        A = sps.coo_matrix((A_data, (A_irow, A_icol)), A_shape)
        
        # build b
        b_data = [1., 0.]
        
        # calc
        toc = time.perf_counter()
        res = _lp_solver(self.method, c_data, G, h_data, A, b_data)
        self.time_level2 = time.perf_counter() - toc
 
        self.status = res['status']
        if self.status != 0:
           warnings.warn(f"Warning {self.name} on {self.rrate.index[-1]} :: "
                         f"status {res['status']} :: {res['infostring']}")
           return np.array([np.nan] * mm)
            
        t = res['x'][-1]
        # Gini-Divers
        self.diverse= 1. - res['pcost']
        # Gini
        self.risk = res['pcost'] / t
        # optimal weights
        self.ww = np.array(res['x'][:mm] / t)
        self.ww.shape = mm
        # rate of return
        self.RR = np.dot(self.ww, self.muk)
        # default to risk
        self.primary_risk_comp = np.array([self.risk])
        self.secondary_risk_comp = np.array([self.risk])
        
        return self.ww
    
    
    def _risk_inv_diversification(self, d=1):
        # Order of variables (mm - no of symb, nn - no of observations)
        # w <- [0 : mm]
        # s <- [mm : mm + nn2]
        # t <- mm + nn2
        # in total dim = mm + nn2 + 1
        # where nn2 = nn * (nn - 1) / 2
        mm = self.mm
        nn2 = self.nn2
        norm = 1 / self.nn ** 2
        
        # build c
        c_data = list(-self.risk_comp) + [0.] * (nn2 + 1)
        
        # build G
        G_icol = [m for m in range(mm) for _ in range(nn2)] * 2
        G_irow = list(range(nn2)) * mm + list(range(nn2, 2 * nn2)) * mm
        G_data = list(self.drate) + list(-self.drate)
        
        G_icol += list(range(mm, mm + nn2)) * 2
        G_irow += list(range(nn2 * 2))
        G_data += [-1.] * (nn2 * 2)
        
        G_icol += list(range(mm)) + [mm + nn2]
        G_irow += [nn2 * 2] * (mm + 1)
        G_data += list(-self.muk * d) + [self.mu * d]
        
        G_icol += list(range(mm))
        G_irow += list(range(nn2 * 2 + 1, nn2 * 2 + mm + 1))
        G_data += [-1.] * mm
        
        G_shape = (nn2 * 2 + mm + 1, mm + nn2 + 1)
        G = sps.coo_matrix((G_data, (G_irow, G_icol)), G_shape)
        
        # build h
        h_data = [0.] * (nn2 * 2 + mm + 1) 
        
        # build A
        A_icol = list(range(mm, mm + nn2))
        A_irow = [0] * nn2
        A_data = [norm] * nn2
        
        A_icol += list(range(mm)) + [mm + nn2]
        A_irow += [1] * (mm + 1)
        A_data += [1.] * mm + [-1.]
        
        A_shape = (2, mm + nn2 + 1)
        A = sps.coo_matrix((A_data, (A_irow, A_icol)), A_shape)
        
        # build b
        b_data = [1., 0.]
        
        # calc
        toc = time.perf_counter()
        res = _lp_solver(self.method, c_data, G, h_data, A, b_data)
        self.time_level2 = time.perf_counter() - toc
 
        self.status = res['status']
        if self.status != 0:
            warnings.warn(f"Warning {self.name} on {self.rrate.index[-1]} :: "
                          f"status {res['status']} :: {res['infostring']}")
            return np.array([np.nan] * mm)
            
        t = res['x'][-1]
        # Gini-Divers
        self.diverse= 1. + 1. / res['pcost']
        # Gini
        self.risk = 1. / t
        # optimal weights
        self.ww = np.array(res['x'][:mm] / t)
        self.ww.shape = mm
        # rate of return
        self.RR = np.dot(self.ww, self.muk)
        # default to risk
        self.primary_risk_comp = np.array([self.risk])
        self.secondary_risk_comp = np.array([self.risk])
        
        return self.ww
        

    def _rr_max_diversification(self):
        # Order of variables (mm - no of symb, nn - no of observations)
        # w <- [0 : mm]
        # s <- [mm : mm + nn2]
        # in total dim = mm + nn2
        # where nn2 = nn * (nn - 1) / 2
        mm = self.mm
        nn2 = self.nn2
        norm = 1 / self.nn ** 2
        
        # build c
        c_data = list(-self.muk) + [0.] * nn2
        
        # build G
        G_icol = [m for m in range(mm) for _ in range(nn2)] * 2
        G_irow = list(range(nn2)) * mm + list(range(nn2, 2 * nn2)) * mm
        G_data = list(self.drate) + list(-self.drate)
        
        G_icol += list(range(mm, mm + nn2)) * 2
        G_irow += list(range(nn2 * 2))
        G_data += [-1.] * (nn2 * 2)
        
        G_icol += list(range(mm))
        G_irow += list(range(nn2 * 2, nn2 * 2 + mm))
        G_data += [-1.] * mm
        
        G_shape = (nn2 * 2 + mm, mm + nn2)
        G = sps.coo_matrix((G_data, (G_irow, G_icol)), G_shape)
        
        # build h
        h_data = [0.] * (nn2 * 2 + mm) 
        
        # build A
        A_icol = list(range(mm)) + list(range(mm, mm + nn2)) + list(range(mm))
        A_irow = [0] * mm + [1] * (nn2 + mm)
        A_data = [1.] * mm + [norm] * nn2 + list((self.diverse- 1.) * self.risk_comp)
        
        A_shape = (2, mm + nn2)
        A = sps.coo_matrix((A_data, (A_irow, A_icol)), A_shape)
        
        # build b
        b_data = [1., 0.]
        
        # calc
        toc = time.perf_counter()
        res = _lp_solver(self.method, c_data, G, h_data, A, b_data)
        self.time_level2 = time.perf_counter() - toc
 
        self.status = res['status']
        if self.status != 0:
            warnings.warn(f"Warning {self.name} on {self.rrate.index[-1]} :: "
                          f"status {res['status']} :: {res['infostring']}")
            return np.array([np.nan] * mm)
            
        # optimal weights
        self.ww = np.array(res['x'][:mm])
        self.ww.shape = mm
        # rate of return
        self.RR = -res['pcost']
        # risk
        self.risk = norm * np.sum(res['x'][mm:])
        self.primary_risk_comp = np.array([self.risk])
        self.secondary_risk_comp = np.array([self.risk])
        # diversification
        self.diverse= 1. - self.risk / np.dot(self.ww, self.risk_comp)
        
        return self.ww